#!/usr/bin/env python
#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""This test runner compiles all extension modules for standalone mode.

This is a test to reveal hidden dependencies on a system.

"""

import os
import sys

# Find nuitka package relative to us.
sys.path.insert(
    0,
    os.path.normpath(
        os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", "..")
    ),
)

# isort:start

import shutil

from nuitka.reports.CompilationReportReader import (
    getCompilationOutputBinary,
    parseCompilationReport,
)
from nuitka.tools.testing.Common import (
    check_output,
    checkLoadedFileAccesses,
    checkSucceedsWithCPython,
    compileLibraryTest,
    createSearchMode,
    displayFileContents,
    displayFolderContents,
    displayRuntimeTraces,
    getTempDir,
    my_print,
    setup,
    test_logger,
)
from nuitka.tools.testing.RuntimeTracing import (
    doesSupportTakingRuntimeTrace,
    getRuntimeTraceOfLoadedFiles,
)
from nuitka.utils.Execution import NuitkaCalledProcessError
from nuitka.utils.FileOperations import (
    getFileContents,
    openTextFile,
    withDirectoryChange,
)
from nuitka.utils.ModuleNames import ModuleName
from nuitka.utils.Utils import isMacOS


def displayError(dirname, filename):
    assert dirname is None

    dist_path = filename[:-3] + ".dist"
    displayFolderContents("dist folder", dist_path)

    inclusion_log_path = filename[:-3] + ".py.inclusion.log"
    displayFileContents("inclusion log", inclusion_log_path)


# TODO: Should not have matplotlib, msgpack here, find out why it is.
ignore_packages = ("matplotlib", "black", "Cython", "msgpack", "numba")

if isMacOS():
    # Not supported on macOS
    ignore_packages += ("PyQt6",)


_done_modules = set()


def decide(root, filename):
    for ignore_package in ignore_packages:
        if (
            root.endswith(os.path.sep + ignore_package)
            or os.path.sep + ignore_package + os.path.sep in root
        ):
            return False

    # This is a DLL only
    if filename.endswith("linux-gnu_d.so"):
        return False

    first_part = filename.split(".")[0]
    if first_part in _done_modules:
        return False
    _done_modules.add(first_part)

    return filename.endswith((".so", ".pyd")) and not filename.startswith("libpython")


def _writeImporterTestFile(filename, module_name):
    with openTextFile(filename, "w") as output:
        plugin_names = set(["pylint-warnings"])
        if module_name.hasNamespace("PySide2"):
            plugin_names.add("pyside2")
        elif module_name.hasNamespace("PySide6"):
            plugin_names.add("pyside6")
        elif module_name.hasNamespace("PyQt5"):
            plugin_names.add("pyqt5")
        elif module_name.hasNamespace("PyQt6"):
            plugin_names.add("pyqt6")
        else:
            plugin_names.add("no-qt")

        for plugin_name in plugin_names:
            output.write("# nuitka-project: --enable-plugin=%s\n" % plugin_name)

        # Make it an error to find unwanted bloat compiled in.
        output.write("# nuitka-project: --noinclude-default-mode=error\n")

        # TODO: This won't be bloat for long anymore.
        output.write("# nuitka-project: --noinclude-numba-mode=warning\n")

        output.write("# nuitka-project: --mode=standalone\n")

        if isMacOS():
            output.write("# nuitka-project: --macos-create-app-bundle\n")

            if module_name.hasNamespace("PyQt5"):
                output.write("# nuitka-project: --mode=onefile\n")

        output.write("import " + module_name.asString() + "\n")
        output.write("print('OK.')")


def _checkCompilationResults(
    stage_dir, filename, report_filename, search_mode, command, output
):
    compilation_report = parseCompilationReport(report_filename)

    binary_filename = getCompilationOutputBinary(
        compilation_report=compilation_report,
        prefixes=(("${cwd}", os.getcwd()),),
    )

    outside_accesses = []

    if doesSupportTakingRuntimeTrace():
        with withDirectoryChange(stage_dir):
            loaded_filenames = getRuntimeTraceOfLoadedFiles(
                logger=test_logger,
                command=[binary_filename],
            )

        outside_accesses = checkLoadedFileAccesses(
            loaded_filenames=loaded_filenames, current_dir=stage_dir
        )
    else:
        test_logger.info("Runtime traces are not possible on this machine.")

    if outside_accesses:
        displayError(None, filename)
        displayRuntimeTraces(test_logger, binary_filename)

        test_logger.warning("Should not access these file(s): '%r'." % outside_accesses)

        search_mode.onErrorDetected(1)

    if output[-1] != b"OK.":
        my_print(" ".join(command))
        my_print(filename)
        my_print(output)
        test_logger.sysexit("FAIL.")

    my_print("OK.")

    assert not outside_accesses, outside_accesses

    if os.path.exists(filename[:-3] + ".dist"):
        shutil.rmtree(filename[:-3] + ".dist")


def main():
    setup(suite="extension_modules", needs_io_encoding=True)
    search_mode = createSearchMode()

    tmp_dir = getTempDir()

    def action(stage_dir, root, path):
        report_filename = "test-compilation-report.xml"

        command = [
            sys.executable,
            os.path.join("..", "..", "bin", "nuitka"),
            "--stand",
            "--run",
            "--output-dir=%s" % stage_dir,
            "--remove-output",
            "--no-progressbar",
            "--report=%s" % report_filename,
        ]

        filename = os.path.join(stage_dir, "importer.py")

        assert path.startswith(root)

        module_name = path[len(root) + 1 :]
        module_name = module_name.split(".")[0]
        module_name = module_name.replace(os.path.sep, ".")

        module_name = ModuleName(module_name)

        _writeImporterTestFile(filename, module_name)

        command += os.environ.get("NUITKA_EXTRA_OPTIONS", "").split()

        command.append(filename)

        if checkSucceedsWithCPython(filename):
            try:
                output = check_output(command).splitlines()
            except NuitkaCalledProcessError as e:
                my_print("SCRIPT:", filename, style="blue")
                my_print(getFileContents(filename))

                test_logger.sysexit("Error with compilation: %s" % e)
            # only trying to check for no exception, pylint: disable=try-except-raise
            except Exception:
                raise
            else:
                _checkCompilationResults(
                    stage_dir=stage_dir,
                    filename=filename,
                    report_filename=report_filename,
                    search_mode=search_mode,
                    command=command,
                    output=output,
                )
        else:
            my_print("SKIP (does not work with CPython)")

    compileLibraryTest(
        search_mode=search_mode,
        stage_dir=os.path.join(tmp_dir, "compile_extensions"),
        decide=decide,
        action=action,
    )

    my_print("FINISHED, all extension modules compiled.")


if __name__ == "__main__":
    main()

#     Python test originally created or extracted from other peoples work. The
#     parts from me are licensed as below. It is at least Free Software where
#     it's copied from other people. In these cases, that will normally be
#     indicated.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#!/usr/bin/env python
#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Main program for auto format tool."""

from nuitka.options.CommandLineOptionsTools import makeOptionsParser
from nuitka.Progress import enableProgressBar, wrapWithProgressBar
from nuitka.tools.quality.auto_format.AutoFormat import autoFormatFile
from nuitka.tools.quality.Git import (
    addGitArguments,
    getCheckoutFileChangeDesc,
    getGitPaths,
)
from nuitka.tools.quality.ScanSources import scanTargets
from nuitka.Tracing import my_print, tools_logger
from nuitka.utils.FileOperations import resolveShellPatternToFilenames


def _addOptions(parser):
    parser.add_option(
        "--verbose",
        action="store_true",
        dest="verbose",
        default=False,
        help="""Default is %default.""",
    )

    addGitArguments(parser, verb="Auto-format")

    parser.add_option(
        "--from-commit",
        action="store_true",
        dest="from_commit",
        default=False,
        help="""From commit hook, do not descend into directories. Default is %default.""",
    )

    parser.add_option(
        "--check-only",
        action="store_true",
        dest="check_only",
        default=False,
        help="""For CI testing, check if it's properly formatted. Default is %default.""",
    )

    parser.add_option(
        "--no-progressbar",
        action="store_false",
        dest="progress_bar",
        default=True,
        help="""Disable progress bar outputs (if tqdm is installed).
Defaults to off.""",
    )

    parser.add_option(
        "--assume-yes-for-downloads",
        action="store_true",
        dest="assume_yes_for_downloads",
        default=False,
        help="""Allow download and execution of tools if needed. Default is %default.""",
    )

    parser.add_option(
        "--yaml",
        action="store_true",
        dest="yaml",
        default=False,
        help="""Format only matching Yaml files
Defaults to off.""",
    )

    parser.add_option(
        "--python",
        action="store_true",
        dest="python",
        default=False,
        help="""Format only matching Python files
Defaults to off.""",
    )

    parser.add_option(
        "--c",
        action="store_true",
        dest="c",
        default=False,
        help="""Format only matching C files
Defaults to off.""",
    )

    parser.add_option(
        "--rst",
        action="store_true",
        dest="rst",
        default=False,
        help="""Format only matching rst files
Defaults to off.""",
    )

    parser.add_option(
        "--md",
        action="store_true",
        dest="md",
        default=False,
        help="""Format only matching markdown files
Defaults to off.""",
    )

    parser.add_option(
        "--json",
        action="store_true",
        dest="json",
        default=False,
        help="""Format only matching JSON files
Defaults to off.""",
    )


def _parseArgs():
    parser = makeOptionsParser(usage=None, epilog=None)

    _addOptions(parser)

    options, positional_args = parser.parse_args()

    if options.diff or options.un_pushed:
        if options.from_commit:
            return tools_logger.sysexit(
                "Error, no --from-commit argument allowed in git diff mode."
            )

    if options.progress_bar:
        enableProgressBar(progress_bar="auto")

    return options, positional_args


def formatFilenames(filenames, options):
    result = 0

    for filename in wrapWithProgressBar(filenames, stage="Auto format", unit="file"):
        try:
            if autoFormatFile(
                filename,
                git_stage=False,
                check_only=options.check_only,
                limit_yaml=options.yaml,
                limit_c=options.c,
                limit_python=options.python,
                limit_rst=options.rst,
                limit_md=options.md,
                limit_json=options.json,
                assume_yes_for_downloads=options.assume_yes_for_downloads,
            ):
                result += 1
        except Exception as e:
            tools_logger.warning("Error formatting %s: %s" % (filename, e))
            raise

    # Tool is named without separator, spellchecker: ignore autoformat

    if options.check_only and result > 0:
        return tools_logger.sysexit(
            """Error, 'bin/autoformat-nuitka-source' would make changes to %d files, \
make sure to have commit hook installed or run it manually.""" % result
        )
    elif result > 0:
        tools_logger.info("autoformat: Changes to formatting of %d files" % result)
    else:
        tools_logger.info("autoformat: No files needed formatting changes.")


def _formatFromCommit(options):
    for git_stage in getCheckoutFileChangeDesc(staged=True):
        autoFormatFile(
            git_stage["src_path"],
            git_stage=git_stage,
            assume_yes_for_downloads=options.assume_yes_for_downloads,
        )


def _formatFromGitPaths(options, positional_args):
    positional_args = getGitPaths(
        options=options,
        positional_args=positional_args,
        default_positional_args=(
            "bin",
            "lib",
            "misc",
            "nuitka",
            "rpm",
            "setup.py",
            "tests",
            ".github",
        ),
    )

    if not positional_args:
        return tools_logger.sysexit("No files found.")

    if options.check_only:
        my_print("Checking:", ", ".join(positional_args))
    else:
        my_print("Working on:", ", ".join(positional_args))

    positional_args = sum(
        (
            resolveShellPatternToFilenames(positional_arg)
            for positional_arg in positional_args
        ),
        [],
    )

    filenames = list(
        scanTargets(
            positional_args,
            suffixes=(
                ".py",
                ".scons",
                ".rst",
                ".txt",
                ".j2",
                ".md",
                ".c",
                ".h",
                ".yml",
                ".json",
                ".cursorrules",
            ),
        )
    )
    if options.verbose:
        my_print("Selected:", ", ".join(filenames))

    if not filenames:
        return tools_logger.sysexit("No files found.")

    formatFilenames(filenames, options)


def main():
    options, positional_args = _parseArgs()

    if options.from_commit:
        _formatFromCommit(options)
    else:
        _formatFromGitPaths(options, positional_args)


if __name__ == "__main__":
    main()

#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        https://www.gnu.org/licenses/agpl-3.0.txt
#
#     See also: "Nuitka Runtime Library Exception, Version 1.0" in file
#     "LICENSE-RUNTIME.txt" for additional permissions granted under Section 7.
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


# Version: 0.28
# Heavily truncated version of versioneer

import configparser
import json
import os
import sys
from pathlib import Path

have_tomllib = True
if sys.version_info >= (3, 11):
    import tomllib
else:
    try:
        import tomli as tomllib
    except ImportError:
        have_tomllib = False


class VersioneerConfig:
    """Container for Versioneer configuration parameters."""


def get_root():
    """Get the project root directory.

    We require that all commands are run from the project root, i.e. the
    directory that contains setup.py, setup.cfg, and versioneer.py .
    """
    root = os.path.realpath(os.path.abspath(os.getcwd()))
    setup_py = os.path.join(root, "setup.py")
    versioneer_py = os.path.join(root, "versioneer.py")
    if not (os.path.exists(setup_py) or os.path.exists(versioneer_py)):
        # allow 'python path/to/setup.py COMMAND'
        root = os.path.dirname(os.path.realpath(os.path.abspath(sys.argv[0])))
        setup_py = os.path.join(root, "setup.py")
        versioneer_py = os.path.join(root, "versioneer.py")
    if not (os.path.exists(setup_py) or os.path.exists(versioneer_py)):
        err = (
            "Versioneer was unable to run the project root directory. "
            "Versioneer requires setup.py to be executed from "
            "its immediate directory (like 'python setup.py COMMAND'), "
            "or in a way that lets it use sys.argv[0] to find the root "
            "(like 'python path/to/setup.py COMMAND')."
        )
        raise VersioneerBadRootError(err)
    return root


def get_config_from_root(root):
    """Read the project setup.cfg file to determine Versioneer config."""
    # This might raise OSError (if setup.cfg is missing)

    root = Path(root)
    pyproject_toml = root / "pyproject.toml"
    setup_cfg = root / "setup.cfg"
    section = None
    if pyproject_toml.exists() and have_tomllib:
        try:
            with open(pyproject_toml, "rb") as fobj:
                pp = tomllib.load(fobj)
            section = pp["tool"]["versioneer"]
        except (tomllib.TOMLDecodeError, KeyError):
            pass
    if not section:
        parser = configparser.ConfigParser()
        with open(setup_cfg) as cfg_file:
            parser.read_file(cfg_file)

        section = parser["versioneer"]

    cfg = VersioneerConfig()
    cfg.versionfile_source = section.get("versionfile_source")
    cfg.versionfile_build = section.get("versionfile_build")

    cfg.verbose = section.get("verbose")
    return cfg


SHORT_VERSION_PY = """
# This file was generated by 'versioneer.py' (0.28) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json

version_json = '''
%s
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)
"""


def write_to_version_file(filename, versions):
    """Write the given version number to the given _version.py file."""
    os.unlink(filename)
    contents = json.dumps(versions, sort_keys=True, indent=1, separators=(",", ": "))
    with open(filename, "w") as f:
        f.write(SHORT_VERSION_PY % contents)

    print("set %s to '%s'" % (filename, versions["version"]))


class VersioneerBadRootError(Exception):
    """The project root directory is unknown or missing key files."""


def get_versions(verbose=False):
    """Get the project version from whatever source is available.

    Returns dict with two keys: 'version' and 'full'.
    """
    if "versioneer" in sys.modules:
        # see the discussion in cmdclass.py:get_cmdclass()
        del sys.modules["versioneer"]

    # We return a special version always for testing
    return {
        "version": "0+24601",
        "full-revisionid": None,
        "dirty": None,
        "error": "unable to compute version",
        "date": None,
    }


def get_version():
    """Get the short version string for this project."""
    return get_versions()["version"]


def get_cmdclass(cmdclass=None):
    """Get the custom setuptools subclasses used by Versioneer.

    If the package uses a different cmdclass (e.g. one from numpy), it
    should be provide as an argument.
    """
    if "versioneer" in sys.modules:
        del sys.modules["versioneer"]
        # this fixes the "python setup.py develop" case (also 'install' and
        # 'easy_install .'), in which subdependencies of the main project are
        # built (using setup.py bdist_egg) in the same python process. Assume
        # a main project A and a dependency B, which use different versions
        # of Versioneer. A's setup.py imports A's Versioneer, leaving it in
        # sys.modules by the time B's setup.py is executed, causing B to run
        # with the wrong versioneer. Setuptools wraps the sub-dep builds in a
        # sandbox that restores sys.modules to it's pre-build state, so the
        # parent is protected against the child's "import versioneer". By
        # removing ourselves from sys.modules here, before the child build
        # happens, we protect the child from the parent's versioneer too.
        # Also see https://github.com/python-versioneer/python-versioneer/issues/52

    cmds = {} if cmdclass is None else cmdclass.copy()

    # we override "build_py" in setuptools
    #
    # most invocation pathways end up running build_py:
    #  distutils/build -> build_py
    #  distutils/install -> distutils/build ->..
    #  setuptools/bdist_wheel -> distutils/install ->..
    #  setuptools/bdist_egg -> distutils/install_lib -> build_py
    #  setuptools/install -> bdist_egg ->..
    #  setuptools/develop -> ?
    #  pip install:
    #   copies source tree to a tempdir before running egg_info/etc
    #   if .git isn't copied too, 'git describe' will fail
    #   then does setup.py bdist_wheel, or sometimes setup.py install
    #  setup.py egg_info -> ?

    # pip install -e . and setuptool/editable_wheel will invoke build_py
    # but the build_py command is not expected to copy any files.

    # we override different "build_py" commands for both environments
    if "build_py" in cmds:
        _build_py = cmds["build_py"]
    else:
        from setuptools.command.build_py import build_py as _build_py

    class cmd_build_py(_build_py):
        def run(self):
            root = get_root()
            cfg = get_config_from_root(root)
            versions = get_versions()
            _build_py.run(self)
            if getattr(self, "editable_mode", False):
                # During editable installs `.py` and data files are
                # not copied to build_lib
                return
            # now locate _version.py in the new build/ directory and replace
            # it with an updated value
            if cfg.versionfile_build:
                target_versionfile = os.path.join(self.build_lib, cfg.versionfile_build)
                print("UPDATING %s" % target_versionfile)
                write_to_version_file(target_versionfile, versions)

    cmds["build_py"] = cmd_build_py

    return cmds


#     Python test originally created or extracted from other peoples work. The
#     parts from me are licensed as below. It is at least Free Software where
#     it's copied from other people. In these cases, that will normally be
#     indicated.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#!/usr/bin/python
#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Small tool to create a pbuilder image for Nuitka private CI mainly."""

import os
import shutil
import subprocess
import sys
import tempfile

debian, codename, output = sys.argv[1:]

if "-" in codename:
    codename, arch = codename.split("-")
else:
    arch = subprocess.check_output("dpkg-architecture -q DEB_HOST_ARCH".split()).strip()

    if str is not bytes:
        arch = arch.decode("utf8")

start_dir = os.getcwd()

stage = tempfile.mkdtemp()

try:
    os.chdir(stage)

    if debian == "debian":
        # Certificate for https is not good.
        mirror = "http://ftp.us.debian.org/debian"
        components = "main"
    elif debian == "ubuntu":
        mirror = "http://de.archive.ubuntu.com/ubuntu"
        components = "main,universe"
    else:
        assert False, debian

    subprocess.check_call(
        [
            "debootstrap",  # spell-checker: ignore debootstrap
            "--include=ccache",
            "--include=dpkg-dev",
            "--arch=" + arch,
            "--components=" + components,
            codename,
            "chroot",
            mirror,
        ]
    )

    subprocess.check_output(["rm", "-rf", "chroot/var/cache/apt/archives"])

    os.makedirs("chroot/var/cache/apt/archives")

    os.makedirs("chroot/etc/apt.conf.d")
    with open("chroot/etc/apt.conf.d/75mine", "w", encoding="utf8") as output_file:
        output_file.write('Acquire::Languages "none";\n')

    target_filename = codename + ".tgz"
    subprocess.check_call(["tar", "czf", target_filename, "-C", "chroot", "."])

    shutil.copy(target_filename, os.path.join(start_dir, output))
finally:
    os.chdir(start_dir)
    shutil.rmtree(stage)

#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        https://www.gnu.org/licenses/agpl-3.0.txt
#
#     See also: "Nuitka Runtime Library Exception, Version 1.0" in file
#     "LICENSE-RUNTIME.txt" for additional permissions granted under Section 7.
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#!/usr/bin/env python
#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Test runners test

This shows that typical test runners like nose and py.test can be used with
compiled packages.

"""

import os
import sys

# Find nuitka package relative to us.
sys.path.insert(
    0,
    os.path.normpath(
        os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", "..")
    ),
)

# isort:start

import subprocess

from nuitka.tools.testing.Common import getTempDir, my_print, setup
from nuitka.utils.FileOperations import copyTree, openTextFile


def main():
    setup(suite="test-runners")

    os.chdir("subject")

    nuitka_main_path = os.path.join("..", "..", "..", "bin", "nuitka")
    tmp_dir = getTempDir()

    command = [
        os.environ["PYTHON"],
        nuitka_main_path,
        "--enable-plugin=pylint-warnings",
        "--output-dir=%s" % tmp_dir,
        "--include-package=package",
        "--nofollow-import-to=*.tests",
        "--debug",
        "--mode=module",
        "package",
    ]

    result = subprocess.call(command)

    if result != 0:
        sys.exit(result)

    os.makedirs(os.path.join(tmp_dir, "package.ext"))
    copyTree("package", os.path.join(tmp_dir, "package.ext/package"))

    os.chdir(tmp_dir)

    # We compile the package non-closed, so we can smuggle in tests
    # and user code. This is going to be the example code.
    with openTextFile("package.ext/package/user_provided.py", "w") as output:
        # TODO: Maybe assert that the type name of a local function and one from
        # the package are not the same, i.e. we are running inside the compiled
        # package.
        output.write("""
from __future__ import print_function

import package
print("__name__:",    package.__name__)
print("__package__:", package.__package__)
print("__path__:",    package.__path__)
print("__file__:",    package.__file__)
# print("__loader__:",    package.__loader__)

import package.sub_package1
print("__name__:",    package.sub_package1.__name__)
print("__package__:", package.sub_package1.__package__)
print("__path__:",    package.sub_package1.__path__)
print("__file__:",    package.sub_package1.__file__)
# print("__loader__:",    package.sub_package1.__loader__)

import package.sub_package1.tests;
print("__name__:",    package.sub_package1.tests.__name__)
print("__package__:", package.sub_package1.tests.__package__)
print("__path__:",    package.sub_package1.tests.__path__)
print("__file__:",    package.sub_package1.tests.__file__)
# print("__loader__:",    package.sub_package1.tests.__loader__)
""")

    os.makedirs("nose")
    with openTextFile("nose/usage.txt", "w") as output:
        pass

    os.system("find | sort")

    # Inform about the extra path, format is NUITKA_PACKAGE_fullname where
    # dots become "_" and should point to the directory where external code
    # to be loaded will live under. Probably should be an absolute path, but
    # we avoid it here.
    os.environ["NUITKA_PACKAGE_package"] = "./package.ext/package"

    # Lets make sure these to not work. These will be used in the compiled
    # form only.
    for module_path in ("__init__.py", "sub_package1__init__.py"):
        with openTextFile(
            os.path.join("./package.ext/package", module_path), "w"
        ) as output:
            output.write("assert False")

    # Check the compiled package is functional for importing.
    my_print("Running package as basic test:")
    command = [os.environ["PYTHON"], "-c", "import package"]

    result = subprocess.call(command)

    if result != 0:
        sys.exit(result)

    my_print("Running nose tests:")
    # assert os.system(os.environ["PYTHON"] + " -m nose --first-package-wins -s package.sub_package1.tests") == 0

    my_print("Running py.test tests:")
    command = [
        os.environ["PYTHON"],
        "-m",
        "pytest",
        "-v",
        "--pyargs",
        "package.sub_package1.tests",
    ]

    result = subprocess.call(command)

    if result != 0:
        sys.exit(result)


if __name__ == "__main__":
    main()

#     Python test originally created or extracted from other peoples work. The
#     parts from me are licensed as below. It is at least Free Software where
#     it's copied from other people. In these cases, that will normally be
#     indicated.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


# nuitka-skip-unless-imports: tyro

# nuitka-project: --mode=standalone
# nuitka-project: --enable-plugin=source-inclusion
# nuitka-project: --source-inclusion-args="small --help"
# nuitka-project: --source-inclusion-args="large --help"

# spell-checker: disable

import sys
from dataclasses import dataclass, field

import tyro
from tyro import _docstrings


@dataclass
class BaseTrainerConfig(object):
    """Base trainer settings."""

    name: str = "base"
    """Base trainer name."""


@dataclass
class MachineConfig(object):
    """Machine settings."""

    workers: int = 4
    """Number of workers."""


@dataclass
class LoggingConfig(object):
    """Logging settings."""

    output_dir: str = "logs"
    """Output directory for logs."""


@dataclass
class SmallTrainerConfig(BaseTrainerConfig):
    """Small trainer settings."""

    machine: MachineConfig = field(default_factory=MachineConfig)
    logging: LoggingConfig = field(default_factory=LoggingConfig)
    steps: int = 100
    """Number of training steps."""


@dataclass
class LargeTrainerConfig(BaseTrainerConfig):
    """Large trainer settings."""

    machine: MachineConfig = field(default_factory=MachineConfig)
    logging: LoggingConfig = field(default_factory=LoggingConfig)
    precision: str = "fp16"
    """Numeric precision for the large trainer."""


AnnotatedBaseConfigUnion = tyro.conf.SuppressFixed[
    tyro.conf.FlagConversionOff[
        tyro.extras.subcommand_type_from_defaults(
            {
                "small": SmallTrainerConfig(name="small", steps=25),
                "large": LargeTrainerConfig(name="large", precision="bf16"),
            },
            prefix_names=False,
        )
    ]
]


def main():
    args = sys.argv[1:]

    if not args:
        args = ["small"]

    config = tyro.cli(AnnotatedBaseConfigUnion, args=args)

    steps_doc = _docstrings.get_field_docstring(SmallTrainerConfig, "steps", ())
    precision_doc = _docstrings.get_field_docstring(LargeTrainerConfig, "precision", ())
    workers_doc = _docstrings.get_field_docstring(MachineConfig, "workers", ())

    assert steps_doc == "Number of training steps.", steps_doc
    assert precision_doc == "Numeric precision for the large trainer.", precision_doc
    assert workers_doc == "Number of workers.", workers_doc

    print("Trainer:", config.name, type(config).__name__)
    print("Steps doc:", steps_doc)
    print("Precision doc:", precision_doc)
    print("Workers doc:", workers_doc)
    print("OK.")


if __name__ == "__main__":
    main()

#     Python test originally created or extracted from other peoples work. The
#     parts from me are licensed as below. It is at least Free Software where
#     it's copied from other people. In these cases, that will normally be
#     indicated.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Path specifications and templating."""

import os

from nuitka.Tracing import onefile_logger, options_logger
from nuitka.utils.FileOperations import getUserInputNormalizedPath, isLegalPath

from .OptionParsing import run_time_variable_names


def _convertOldStylePathSpecQuotes(value):
    quote = None

    result = ""
    for c in value:
        if c == "%":
            if quote is None:
                quote = "{"
                result += quote
            elif quote == "{":
                result += "}"
                quote = None
        else:
            result += c

    return result


def checkPathSpec(value, arg_name, allow_disable):
    # There are never enough checks here and sysexit is returned,
    # pylint: disable=too-many-branches,too-many-return-statements
    from .Options import (
        getCompanyName,
        getFileVersionTuple,
        getProductName,
        getProductVersionTuple,
    )

    old = value
    value = _convertOldStylePathSpecQuotes(value)
    if old != value:
        options_logger.warning(
            "Adapted '%s' option value from legacy quoting style to '%s' -> '%s'"
            % (arg_name, old, value)
        )

    # This changes the '/' to '\\' on Windows at least.
    value = getUserInputNormalizedPath(value)

    if "\n" in value or "\r" in value:
        return options_logger.sysexit(
            "Using a new line in value '%s=%r' value is not allowed."
            % (arg_name, value)
        )

    if "{NONE}" in value:
        if not allow_disable:
            return options_logger.sysexit(
                "Using value '{NONE}' in '%s=%s' value is not allowed."
                % (arg_name, value)
            )

        if value != "{NONE}":
            return options_logger.sysexit(
                "Using value '{NONE}' in '%s=%s' value does not allow anything else used too."
                % (arg_name, value)
            )

    if "{NULL}" in value:
        if not allow_disable:
            return options_logger.sysexit(
                "Using value '{NULL}' in '%s=%s' value is not allowed."
                % (arg_name, value)
            )

        if value != "{NULL}":
            return options_logger.sysexit(
                "Using value '{NULL}' in '%s=%s' value does not allow anything else used too."
                % (arg_name, value)
            )

    if "{COMPANY}" in value and not getCompanyName():
        return options_logger.sysexit(
            "Using value '{COMPANY}' in '%s=%s' value without '--company-name' being specified."
            % (arg_name, value)
        )

    if "{PRODUCT}" in value and not getProductName():
        return options_logger.sysexit(
            "Using value '{PRODUCT}' in '%s=%s' value without '--product-name' being specified."
            % (arg_name, value)
        )

    if "{VERSION}" in value and not (getFileVersionTuple() or getProductVersionTuple()):
        return options_logger.sysexit(
            "Using value '{VERSION}' in '%s=%s' value without '--product-version' or '--file-version' being specified."
            % (arg_name, value)
        )

    if "{FILE_VERSION}" in value and not getFileVersionTuple():
        return options_logger.sysexit(
            "Using value '{FILE_VERSION}' in '%s=%s' value without '--file-version' being specified."
            % (arg_name, value)
        )

    if "{PRODUCT_VERSION}" in value and not getProductVersionTuple():
        return options_logger.sysexit(
            "Using value '{PRODUCT_VERSION}' in '%s=%s' value without '--product-version' being specified."
            % (arg_name, value)
        )

    if value.count("{") != value.count("}"):
        return options_logger.sysexit("""Unmatched '{}' is wrong for '%s=%s' and may \
definitely not do what you want it to do.""" % (arg_name, value))

    # Catch nested or illegal variable names.
    var_name = None
    for c in value:
        if c in "{":
            if var_name is not None:
                return options_logger.sysexit(
                    """Nested '{' is wrong for '%s=%s'.""" % (arg_name, value)
                )
            var_name = ""
        elif c == "}":
            if var_name is None:
                return options_logger.sysexit(
                    """Stray '}' is wrong for '%s=%s'.""" % (arg_name, value)
                )

            if var_name not in run_time_variable_names:
                return onefile_logger.sysexit(
                    "Found unknown variable name '{%s}' in for '%s=%s'."
                    "" % (var_name, arg_name, value)
                )

            var_name = None
        else:
            if var_name is not None:
                var_name += c

    for candidate in (
        "{PROGRAM}",
        "{PROGRAM_BASE}",
        "{PROGRAM_DIR}",
        "{CACHE_DIR}",
        "{HOME}",
        "{TEMP}",
    ):
        if candidate in value[1:]:
            return options_logger.sysexit("""\
Absolute run time paths of '%s' can only be at the start of \
'%s=%s', using it in the middle of it is not allowed.""" % (candidate, arg_name, value))

        if candidate == value:
            return options_logger.sysexit("""Cannot use folder '%s', may only be the \
start of '%s=%s', using that alone is not allowed.""" % (candidate, arg_name, value))

        if value.startswith(candidate) and candidate != "{PROGRAM_BASE}":
            if value[len(candidate)] != os.path.sep:
                return options_logger.sysexit(
                    """Cannot use general system folder %s, without a path \
separator '%s=%s', just appending to these is not allowed, needs to be \
below them.""" % (candidate, arg_name, value)
                )

    is_legal, reason = isLegalPath(value)
    if not is_legal:
        return options_logger.sysexit(
            """Cannot use illegal paths '%s=%s', due to %s."""
            % (arg_name, value, reason)
        )

    return value


def expandPathSpec(value):
    """Expand templated variables like {VERSION} into their actual string values.

    Notes:
        This evaluates the compile time available variables. Values like
        {PID} are not available and evaluated at run time by the binary.
    """
    from .Options import (
        getCompanyName,
        getFileVersion,
        getProductName,
        getProductVersion,
    )

    if "{VERSION}" in value:

        product_version = getProductVersion()
        file_version = getFileVersion()

        if file_version is not None and product_version is not None:
            effective_version = "%s-%s" % (
                product_version,
                file_version,
            )
        else:
            effective_version = file_version or product_version

        value = value.replace("{VERSION}", effective_version or "")

    if "{FILE_VERSION}" in value:
        version = getFileVersion() or ""
        value = value.replace("{FILE_VERSION}", version)

    if "{PRODUCT_VERSION}" in value:
        version = getProductVersion() or ""
        value = value.replace("{PRODUCT_VERSION}", version)

    if "{COMPANY}" in value:
        company = getCompanyName() or ""
        value = value.replace("{COMPANY}", company)

    if "{PRODUCT}" in value:
        product = getProductName() or ""
        value = value.replace("{PRODUCT}", product)

    return value


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        https://www.gnu.org/licenses/agpl-3.0.txt
#
#     See also: "Nuitka Runtime Library Exception, Version 1.0" in file
#     "LICENSE-RUNTIME.txt" for additional permissions granted under Section 7.
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Helper functions for the Python build related scons files."""

import os

from nuitka.Tracing import scons_logger
from nuitka.utils.Utils import isLinux, isMacOS, isWin32Windows

from .SconsUtils import getArgumentDefaulted, getArgumentRequired, isGccName

# spell-checker: ignore cppdefines,cpppath,linkflags,libpath


def _addPythonIncludePaths(env):
    python_include_paths = []

    adapted_python_header_files_dir = getArgumentDefaulted(
        "adapted_python_header_files_dir", ""
    )
    if adapted_python_header_files_dir:
        python_include_paths.append(adapted_python_header_files_dir)
        env.Append(CPPDEFINES=["_NUITKA_ADAPTED_PYTHON_HEADERS"])

    python_include_path = getArgumentRequired("python_include_path")
    python_include_paths.append(python_include_path)

    if env.python_version >= (3, 13):
        # spell-checker: ignore mimalloc
        python_include_paths.append(
            os.path.join(python_include_path, "internal", "mimalloc")
        )

    if env.self_compiled_python_uninstalled:
        if isWin32Windows():
            python_header_path = os.path.join(env.python_prefix_external, "PC")

            if os.path.exists(python_header_path):
                python_include_paths.append(python_header_path)
        else:
            python_include_paths.append(env.python_prefix_external)

    env.Append(CPPPATH=python_include_paths)


def _addPythonWarningSettings(env):
    if env.debug_mode:
        if env.gcc_mode:
            env.Append(
                CCFLAGS=[
                    # Unfortunately Py_INCREF(Py_False) triggers aliasing warnings,
                    # which are unfounded, so disable them.
                    "-Wno-error=strict-aliasing",
                    "-Wno-strict-aliasing",
                    # At least for self-compiled Python3.2, and MinGW this happens
                    # and has little use anyway.
                    "-Wno-error=format",
                    "-Wno-format",
                ]
            )

            if isGccName(env.the_cc_name):
                # Some Linux distro hardening injects "-Werror=format-security",
                # which would otherwise conflict with disabling format warnings.
                # spell-checker: ignore Werror
                env.Append(
                    CCFLAGS=[
                        "-Wno-error=format-security",
                    ]
                )

        elif env.msvc_mode:
            # Disable warnings that system headers already show.
            env.Append(
                CCFLAGS=[
                    "/W4",
                    "/wd4505",
                    "/wd4127",
                    "/wd4100",
                    "/wd4702",
                    "/wd4189",
                    "/wd4211",
                    "/wd4115",
                ]
            )

            # Disable warnings, that CPython headers already show.
            if env.python_version >= (3,):
                env.Append(CCFLAGS=["/wd4512", "/wd4510", "/wd4610"])

            if env.python_version >= (3, 13):
                env.Append(CCFLAGS=["/wd4324"])

            # We use null arrays in our structure Python declarations, which C11 does
            # not really allow, but should work.
            env.Append(CCFLAGS=["/wd4200"])

            # Do not show deprecation warnings, we will use methods for as long
            # as they work.
            env.Append(CCFLAGS=["/wd4996"])


def applyPythonBuildSettings(env):
    _addPythonIncludePaths(env)
    _addPythonWarningSettings(env)

    if env.monolithpy:
        env.Append(CPPDEFINES=["_MONOLITHPY"])

    if env.static_libpython:
        env.Append(CPPDEFINES=["_NUITKA_STATIC_LIBPYTHON"])

    if env.python_debug:
        env.Append(CPPDEFINES=["Py_DEBUG", "Py_NO_LINK_LIB"])

    if not env.gil_mode:
        env.Append(CPPDEFINES="Py_GIL_DISABLED")

    # We need "dl" in accelerated mode.
    if isLinux():
        env.Append(LIBS=["dl"])

    if not env.msvc_mode:
        env.Append(LIBS=["m"])

    if env.static_libpython:
        env.Append(CPPDEFINES=["Py_NO_ENABLE_SHARED"])

    if env.msvc_mode and (env.module_mode or env.dll_mode):
        # Make sure we handle import library on our own and put it into the
        # build directory, spell-checker: ignore IMPLIB
        env.no_import_lib = True
        env.Append(
            LINKFLAGS=[
                "/IMPLIB:%s" % os.path.join(env.source_dir, "import.lib"),
            ]
        )
    else:
        env.no_import_lib = False

    if env.deployment_mode:
        env.Append(CPPDEFINES=["_NUITKA_DEPLOYMENT_MODE"])

    if env.frozen_modules:
        env.Append(CPPDEFINES=["_NUITKA_FROZEN=%d" % env.frozen_modules])


def addWin32PythonLib(env):
    # Make sure to locate the Python link library from multiple potential
    # locations (installed vs. self-built).
    if env.msys2_mingw_python:
        win_lib_name = "libpython" + env.python_abi_version + ".dll.a"
        win_lib_filename = win_lib_name
    elif env.python_debug:
        win_lib_name = "python" + env.python_abi_version.replace(".", "") + "_d"
        win_lib_filename = win_lib_name + ".lib"
    else:
        win_lib_name = "python" + env.python_abi_version.replace(".", "")
        win_lib_filename = win_lib_name + ".lib"

    if env.python_version >= (3,):
        pc_build_dir = (
            "PCBuild\\amd64" if env.target_arch == "x86_64" else "PCBuild\\win32"
        )
    else:
        pc_build_dir = "PCBuild"

    for candidate in ("libs", "lib", pc_build_dir):
        win_lib_path = os.path.join(env.python_prefix_external, candidate)

        if os.path.exists(os.path.join(win_lib_path, win_lib_filename)):
            break
    else:
        scons_logger.sysexit(
            "Error, cannot find link library '%s' file." % win_lib_filename
        )

    env.Append(LIBPATH=[win_lib_path])
    env.Append(LIBS=[win_lib_name])


def addPythonHaclLib(env, link_module_libs):
    if env.static_libpython and not isMacOS():
        if env.python_version >= (3, 14):
            # Not practical anymore, to build it ourselves.
            hacl_version = None
        elif env.python_version >= (3, 12):
            hacl_version = "hacl_312"
        else:
            hacl_version = None

        if hacl_version is not None:
            env.Append(
                CPPPATH=[
                    os.path.join(
                        env.nuitka_src,
                        "inline_copy",
                        "python_hacl",
                        hacl_version,
                    ),
                    os.path.join(
                        env.nuitka_src,
                        "inline_copy",
                        "python_hacl",
                        hacl_version,
                        "include",
                    ),
                ]
            )

            env.Append(CPPDEFINES=["_NUITKA_INLINE_COPY_HACL"])

            # env.Append(CPPDEFINES=["HACL_CAN_COMPILE_VEC128"])
            env.Append(CPPDEFINES=["HACL_CAN_COMPILE_VEC256"])

    # Remove it from static link libraries unconditionally - libHacl_Hash_SHA2
    # is an internal CPython build artifact, never shipped as a standalone .a file.
    link_module_libs = [
        link_module_lib
        for link_module_lib in link_module_libs
        if "libHacl_Hash_SHA2" not in link_module_lib
    ]

    return link_module_libs


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        https://www.gnu.org/licenses/agpl-3.0.txt
#
#     See also: "Nuitka Runtime Library Exception, Version 1.0" in file
#     "LICENSE-RUNTIME.txt" for additional permissions granted under Section 7.
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#!/usr/bin/env python
#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""OpenSUSE Build Service (OSC) status check.

Confirms that all relevant packages are successfully built. For use in
Buildbot on a timer basis.

"""

import csv
import sys

from nuitka.__past__ import StringIO
from nuitka.tools.environments.Virtualenv import withVirtualenv
from nuitka.Tracing import my_print, tools_logger
from nuitka.utils.InstalledPythons import findInstalledPython


def main():
    # TODO: The setup of the osc virtualenv ought to be shared.
    # many cases, pylint: disable=too-many-branches,too-many-locals,too-many-statements

    my_print("Querying openSUSE build service status of Nuitka packages.")

    # spell-checker: ignore kayhayen
    osc_cmd = "osc pr -c home:kayhayen"

    installed_python = findInstalledPython(
        python_versions=("3.10",),
        module_specs=None,
    )

    with withVirtualenv(
        "venv_nuitka",
        logger=tools_logger,
        style="blue",
        python=installed_python.getPythonExe(),
    ) as venv:
        venv.runCommand("python -m pip install osc")

        stdout_osc, stderr, exit_code = venv.runCommandWithOutput(osc_cmd)

        assert exit_code == 0, stderr

    if str is not bytes:
        stdout_osc = stdout_osc.decode("utf8")

    # Response is really a CSV file, so use that for parsing.
    csv_file = StringIO(stdout_osc)
    osc_reader = csv.reader(csv_file, delimiter=";")

    osc_reader = iter(osc_reader)

    bad = ("failed", "unresolvable", "broken", "blocked")

    titles = next(osc_reader)[1:]

    # Nuitka (follow git main branch)
    row1 = next(osc_reader)
    # Nuitka-Unstable (follow git develop branch)
    row2 = next(osc_reader)
    # Nuitka-Experimental (follow git factory branch)
    row3 = next(osc_reader)

    problems = []

    def decideConsideration(title, status):
        # Ignore other arch builds, they might to not even boot at times.
        # spell-checker: ignore aarch
        if "ppc" in title or "aarch" in title or "arm" in title:
            return False

        # This fails for other reasons often, and is not critical to Nuitka.
        if "openSUSE_Tumbleweed" in title:
            return False

        # Ignore old Fedora and RHEL6 32 bit being blocked.
        if status == "blocked":
            if (
                "Fedora_2" in title
                or "RedHat_RHEL-6/i586" in title
                or "CentOS_CentOS-6/i586" in title
            ):
                return False

        # It makes building visible now, that's not an error of course.
        if status == "building":
            return False

        # It makes need to build visible as well, that too is not an error
        # really.
        if status == "outdated":
            return False

        return True

    for count, title in enumerate(titles):
        status = row1[count + 1]

        if not decideConsideration(title, status):
            continue

        if status in bad:
            problems.append((row1[0], title, status))

    for count, title in enumerate(titles):
        status = row2[count + 1]

        if not decideConsideration(title, status):
            continue

        if status in bad:
            problems.append((row2[0], title, status))

    for count, title in enumerate(titles):
        status = row3[count + 1]

        if not decideConsideration(title, status):
            continue

        if status in bad:
            problems.append((row3[0], title, status))

    if problems:
        my_print("There are problems with:", style="yellow")
        my_print(
            "\n".join("%s: %s (%s)" % problem for problem in problems), style="yellow"
        )

        if any(problem[0] == "Nuitka" for problem in problems):
            my_print(
                "Check here: https://build.opensuse.org/package/show/home:kayhayen/Nuitka"
            )

        if any(problem[0] == "Nuitka-Unstable" for problem in problems):
            my_print(
                "Check here: https://build.opensuse.org/package/show/home:kayhayen/Nuitka-Unstable"
            )

        if any(problem[0] == "Nuitka-experimental" for problem in problems):
            my_print(
                "Check here: https://build.opensuse.org/package/show/home:kayhayen/Nuitka-experimental"
            )

        sys.exit(1)
    else:
        my_print("Looks good.", style="blue")
        sys.exit(0)


if __name__ == "__main__":
    main()

#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        https://www.gnu.org/licenses/agpl-3.0.txt
#
#     See also: "Nuitka Runtime Library Exception, Version 1.0" in file
#     "LICENSE-RUNTIME.txt" for additional permissions granted under Section 7.
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Generate Nuitka CPython offset JSON specialized mappings."""

import os
import re
import sys

from nuitka.build.AdaptPythonHeaderFiles import isOffsetsJsonOutdated
from nuitka.options.CommandLineOptionsTools import makeOptionsParser
from nuitka.tools.Basics import goHome
from nuitka.Tracing import tools_logger
from nuitka.utils.AppDirs import getCacheDir
from nuitka.utils.Download import withUrlOpen
from nuitka.utils.Execution import callProcess, executeProcess
from nuitka.utils.FileOperations import removeDirectory
from nuitka.utils.Utils import getArchitecture, getOS, isMacOS, isWin32Windows


def getInstalledOffsetsJsonPath(python_version, has_gil):
    """
    Return the source distribution bundled JSON storage path resolving strictly against this module.
    The resulting JSON files are permanently checked into version control to cache C struct offsets.
    """
    gil_str = "gil" if has_gil else "no-gil"
    filename = "offsets_%s-%s-%s-%s.json" % (
        python_version,
        getOS(),
        getArchitecture(),
        gil_str,
    )
    return os.path.join(
        os.path.dirname(__file__),
        "..",
        "..",
        "..",
        "build",
        "python_internal_offset",
        filename,
    )


def convertPythonMinorVersionToTuple(v):
    """Convert version strings like '3.13.1' into tuples (3, 13, 1) for precise sorting."""
    assert v.count(".") == 2, v
    return tuple(int(x) for x in v.split("."))


def _getPythonOrgMinorVersions(target_version):
    """
    Scrape the python.org FTP to discover all historically available micro-versions.
    This acts as a completeness backstop because sometimes package managers
    (like uv) do not immediately mirror or cache every single active micro-release.
    """
    tools_logger.info(
        "Querying python.org FTP to ensure latest stable micro-versions...",
        style="tool-progress",
    )

    pattern = re.compile(r'<a href="(%s.\d+)/">' % re.escape(target_version))

    try:
        with withUrlOpen("https://www.python.org/ftp/python/") as response:
            content = response.read().decode("utf8")
    except Exception as e:  # pylint: disable=broad-exception-caught
        return tools_logger.sysexit("Failed to fetch versions from python.org: %s" % e)

    found_versions = set(pattern.findall(content))

    if not found_versions:
        return tools_logger.sysexit(
            "No Python versions found for %s on Python.org." % target_version,
        )

    return found_versions


def _getUvPythonMinorVersions(target_version):
    """
    Exploit 'uv' to list all micro-releases it can natively pull for the target minor.
    Because uv provides an extremely fast, centralized pool of precompiled pythons,
    we prefer mapping as many of these as we can.
    """
    tools_logger.info(
        "Querying uv-Python to get available versions...",
        style="tool-progress",
    )

    found_versions = set()
    result = executeProcess(
        [
            sys.executable,
            "-m",
            "uv",
            "python",
            "list",
            "--only-downloads",
            "--all-versions",
        ]
    )

    if result.exit_code != 0:
        return tools_logger.sysexit("Could not list python versions via uv.")

    output = result.stdout
    if str is not bytes and type(output) is bytes:
        output = output.decode("utf8")

    pattern = re.compile(
        r"cpython-(%s(?:\.\d+)*)-(?:windows|linux|macos)" % re.escape(target_version)
    )
    for line in output.splitlines():
        match = pattern.search(line)
        if match:
            found_versions.add(match.group(1))

    return found_versions


def _getUvPythonBinaryForMinorPythonVersion(python_version_full):
    """
    Install a specific micro-version using 'uv' and return the absolute path to the binary.
    Avoiding manual source compilation speeds up offset generation pipelines by hours,
    which is critical for matrix CI testing.
    """
    tools_logger.info("=> Attempting installation via uv...", style="tool-progress")
    exit_code = callProcess(
        [sys.executable, "-m", "uv", "python", "install", python_version_full]
    )

    if exit_code != 0:
        tools_logger.warning("=> Failed to push installation through uv!")
        return None

    result = executeProcess(
        [sys.executable, "-m", "uv", "python", "find", python_version_full]
    )

    if result.exit_code == 0:
        output = result.stdout
        if str is not bytes and type(output) is bytes:
            output = output.decode("utf8")
        return output.strip()

    return None


def _getPlatformPythonBuildScriptAndExeName():
    """
    Fallback compilation lookup. If precompiled binaries are unavailable (unsupported OS/Arch),
    we must compile CPython entirely from source. This returns the native bootstrap script.
    """
    if isWin32Windows():
        return "compile-python-for-nuitka-windows.cmd", "python.exe"
    elif isMacOS():
        return "compile-python-for-nuitka-mac.sh", os.path.join("bin", "python3")
    else:
        return "compile-python-for-nuitka-linux.sh", os.path.join("bin", "python3")


def _detectAvailablePythonMinorVersions(target_version):
    """
    Unify and sort the combined availability of 'uv' packages and python.org archives,
    ensuring Nuitka extracts offsets for the absolute maximum surface area of releases.
    """
    found_versions = _getUvPythonMinorVersions(
        target_version
    ) | _getPythonOrgMinorVersions(target_version)

    sorted_versions = tuple(
        sorted(found_versions, key=convertPythonMinorVersionToTuple)
    )
    tools_logger.info(
        "Found Python versions: %s" % ", ".join(sorted_versions), style="tool-info"
    )

    return sorted_versions


def _processPythonVersion(python_version_full, options):
    """
    Core orchestration driver for a single micro-version.

    1. Validates if the existing JSON statically checks out to be up-to-date.
    2. Uses 'uv' to provision a pre-built python executable.
    3. Falls back to a heavy CPython source compilation if uv fails.
    4. Recursively invokes Nuitka inside that isolated Python environment
       (--devel-generate-python-internal-offsets) so it can natively build
       and run the specialized C inspection binary locally.
    """
    tools_logger.info(
        "Processing Python %s" % python_version_full, style="tool-progress"
    )

    parts = convertPythonMinorVersionToTuple(python_version_full)
    major = parts[0]
    minor = parts[1]

    json_path = getInstalledOffsetsJsonPath(python_version_full, has_gil=True)

    if isOffsetsJsonOutdated(json_path, "%s.%s" % (major, minor)):
        if options.check:
            tools_logger.warning(
                "=> ERROR: --check mode enabled and mapping for %s is missing!"
                % python_version_full
            )
            return False

        tools_logger.info(
            "=> JSON offset data is missing or outdated.", style="tool-info"
        )
    else:
        tools_logger.info(
            "=> JSON offsets for %s are already up-to-date and complete. Skipping."
            % python_version_full,
            style="tool-success",
        )
        return True

    exe_path_out = _getUvPythonBinaryForMinorPythonVersion(python_version_full)

    if not exe_path_out:
        cache_key = "%s_%s_%s" % (
            getArchitecture(),
            getOS(),
            python_version_full,
        )
        if major > 3 or (major == 3 and minor >= 13):
            cache_key += "_gil"

        python_cache_dir = os.path.join(
            getCacheDir("cpython_source_builds", create=True), cache_key
        )
        script_name, exe_name = _getPlatformPythonBuildScriptAndExeName()

        exe_path_out = os.path.join(python_cache_dir, exe_name)

        if not os.path.exists(exe_path_out):
            tools_logger.info(
                "=> Falling back to native compilation via Python.org source archive!",
                style="tool-fallback",
            )
            compile_cmd = os.path.join("bin", script_name)
            tools_logger.info(
                "=> Launching build sequence. Warning: This takes significant time.",
                style="tool-heavy",
            )

            exit_code = callProcess(
                [
                    compile_cmd,
                    "--version",
                    python_version_full,
                    "--prefix",
                    python_cache_dir,
                    "--cleanup",
                ]
            )

            if exit_code != 0:
                tools_logger.warning(
                    "=> ERROR: Source compilation fallback failed for %s!"
                    % python_version_full
                )
                return False

            assert os.path.exists(exe_path_out), exe_path_out

    if exe_path_out and os.path.exists(exe_path_out):
        tools_logger.info(
            "=> Launching Nuitka offset generation using runtime: %s" % exe_path_out,
            style="tool-progress",
        )
        exit_code = callProcess(
            [
                exe_path_out,
                "-m",
                "nuitka",
                "--devel-generate-python-internal-offsets",
            ]
        )
        if exit_code != 0:
            tools_logger.warning(
                "=> ERROR: Offset generation aborted natively for %s!"
                % python_version_full
            )
            return False

        tools_logger.info(
            "=> Successfully updated offsets for %s" % python_version_full,
            style="tool-success",
        )
    else:
        tools_logger.warning("=> FATAL: Could not locate Python executable!")
        return False

    return True


def _parseOptions():
    parser = makeOptionsParser(
        usage="Usage: %prog [options] [python_version]",
        epilog="Generate or update CPython internal structure information as offset JSON file.",
    )

    parser.add_option(
        "--check",
        action="store_true",
        dest="check",
        default=False,
        help="Only verify that JSONs strictly exist, failing if any releases are missing.",
    )

    parser.add_option(
        "--clean-cache",
        action="store_true",
        dest="clean_cache",
        default=False,
        help="Remove cached python source builds before compilation.",
    )

    options, args = parser.parse_args()

    if args:
        target_version = args[0]
    else:
        target_version = "%s.%s" % (sys.version_info[0], sys.version_info[1])

    if not re.match(r"^3\.\d+$", target_version):
        return tools_logger.sysexit(
            "Error, target version must be of the form '3.x', e.g. '3.13'."
        )

    return options, target_version


def main():
    goHome()

    options, target_version = _parseOptions()

    if options.clean_cache:
        tools_logger.info("Cleaning entire Python source builds cache...")

        removeDirectory(
            path=getCacheDir("cpython_source_builds"),
            logger=tools_logger,
            ignore_errors=True,
            extra_recommendation=None,
        )

    missing_versions = []

    for python_version_full in _detectAvailablePythonMinorVersions(target_version):
        if not _processPythonVersion(python_version_full, options):
            missing_versions.append(python_version_full)

    if missing_versions:
        return tools_logger.sysexit(
            "FATAL: Processing failed or mapped versions missing for: %s"
            % ", ".join(missing_versions)
        )

    if options.check:
        tools_logger.info(
            "All expected micro-versions are natively matched in the workspace keys.",
            style="tool-success",
        )
    else:
        tools_logger.info("Extraction pass natively completed!", style="tool-success")


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        https://www.gnu.org/licenses/agpl-3.0.txt
#
#     See also: "Nuitka Runtime Library Exception, Version 1.0" in file
#     "LICENSE-RUNTIME.txt" for additional permissions granted under Section 7.
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#!/usr/bin/python -u
#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Make PyPI upload of Nuitka, and check success of it."""

import os
import shutil

from nuitka.options.CommandLineOptionsTools import makeOptionsParser
from nuitka.tools.environments.Virtualenv import withVirtualenv
from nuitka.tools.release.Documentation import checkReleaseDocumentation
from nuitka.tools.release.Release import (
    checkBranchName,
    makeNuitkaSourceDistribution,
)
from nuitka.Tracing import tools_logger
from nuitka.utils.InstalledPythons import findInstalledPython
from nuitka.Version import getNuitkaVersion


def _checkNuitkaInVirtualenv(python):
    with withVirtualenv(
        "venv_nuitka", logger=tools_logger, style="blue", python=python.getPythonExe()
    ) as venv:
        tools_logger.info("Installing Nuitka into virtualenv:", style="blue")
        tools_logger.info("*" * 40, style="blue")
        venv.runCommand("python -m pip install ../dist/nuitka*.tar.gz")
        tools_logger.info("*" * 40, style="blue")

        tools_logger.info("Compiling basic test with runner:", style="blue")
        tools_logger.info("*" * 40, style="blue")
        venv.runCommand(
            (
                "nuitka%s ../tests/basics/AssertsTest.py"
                % ("2" if python.getPythonVersion()[0] == "2" else "")
            ),
            style="blue",
        )
        tools_logger.info("*" * 40, style="blue")

        tools_logger.info(
            "Compiling basic test with recommended -m mode:", style="blue"
        )
        tools_logger.info("*" * 40, style="blue")
        venv.runCommand(
            "python -m nuitka ../tests/basics/AssertsTest.py",
            style="blue",
        )
        tools_logger.info("*" * 40, style="blue")


def main():
    nuitka_version = getNuitkaVersion()

    branch_name = checkBranchName()

    parser = makeOptionsParser(usage=None, epilog=None)

    parser.add_option(
        "--token",
        action="store",
        dest="token",
        help="""
Token to use for upload.
""",
    )

    parser.add_option(
        "--check",
        action="store_true",
        dest="check",
        help="""
Check if it would build, without uploading.
""",
    )

    options, positional_args = parser.parse_args()

    if positional_args:
        tools_logger.sysexit(
            "This command takes no positional arguments, check help output."
        )

    # Only real main releases so far.
    if not options.check:
        assert branch_name == "main", branch_name
        assert "pre" not in nuitka_version and "rc" not in nuitka_version

    tools_logger.info("Working on Nuitka %r." % nuitka_version, style="blue")

    shutil.rmtree("check_nuitka", ignore_errors=True)
    shutil.rmtree("dist", ignore_errors=True)

    tools_logger.info("Creating documentation.", style="blue")
    checkReleaseDocumentation()
    tools_logger.info("Creating source distribution.", style="blue")

    # spell-checker: ignore gztar
    dist_filenames = makeNuitkaSourceDistribution(
        formats=("gztar",),
        sign=False,
    )

    # Test with these Pythons if the installed package would work.
    pythons = [
        findInstalledPython(
            python_versions=("2.7",),
            module_specs=None,
        ),
        findInstalledPython(
            python_versions=("3.10",),
            module_specs=None,
        ),
    ]

    for python in pythons:
        _checkNuitkaInVirtualenv(python)

    assert os.system("twine check %s" % dist_filenames[0]) == 0

    if not options.check:
        tools_logger.info("Uploading source dist")

        assert (
            os.system(
                "twine upload --username=__token__ --password=%s %s"
                % (
                    options.token,
                    dist_filenames[0],
                )
            )
            == 0
        )
        tools_logger.info("Uploaded.")
    else:
        tools_logger.info("Checked OK, but not uploaded.")


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        https://www.gnu.org/licenses/agpl-3.0.txt
#
#     See also: "Nuitka Runtime Library Exception, Version 1.0" in file
#     "LICENSE-RUNTIME.txt" for additional permissions granted under Section 7.
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Standard plug-in to make multiprocessing and joblib work well.

On Windows, the multiprocessing modules forks new processes which then have
to start from scratch. This won't work if there is no "sys.executable" to
point to a "Python.exe" and won't use compiled code by default.

The issue applies to accelerated and standalone mode alike.

spell-checker: ignore joblib,anyio
"""

from nuitka.importing.FakeModules import makeFakeModuleDescription
from nuitka.ModuleRegistry import (
    getModuleInclusionInfoByName,
    getRootTopModule,
)
from nuitka.options.Options import isStandaloneMode, shallMakeModule
from nuitka.plugins.PluginBase import NuitkaPluginBase
from nuitka.PythonVersions import python_version
from nuitka.tree.SourceHandling import readSourceCodeFromFilename
from nuitka.utils.ModuleNames import ModuleName


class NuitkaPluginMultiprocessingWorkarounds(NuitkaPluginBase):
    """This is to make multiprocessing work with Nuitka and use compiled code.

    When running in accelerated mode, it's not good to fork a new Python
    instance to run other code, as that won't be accelerated. And when
    run in standalone mode, there may not even be a Python, but it's the
    same principle.

    So by default, this module is on and works around the behavior of the
    "multiprocessing.forking/multiprocessing.spawn/multiprocessing.manager"
    expectations.
    """

    plugin_name = "multiprocessing"
    plugin_desc = "Required by 'multiprocessing' package."
    plugin_category = "package-support"

    def __init__(self):
        self.created_parents_main = False

    @classmethod
    def isRelevant(cls):
        return not shallMakeModule()

    @staticmethod
    def isAlwaysEnabled():
        return True

    @staticmethod
    def createPreModuleLoadCode(module):
        full_name = module.getFullName()

        if full_name == "__parents_main__":
            yield (
                """\
import sys
# If the multiprocessing fork flag is present, scrub the arguments so that the
# user code (which might use argparse) doesn't see them. Save them in builtins
# so we can restore them later when calling multiprocessing.spawn.spawn_main.
if "--multiprocessing-fork" in sys.argv:
    # Use builtins to store the args so they are available globally even in functions
    # without needing to pass them around or use global keyword everywhere.
    if str is bytes:
        import __builtin__ as builtins
    else:
        import builtins

    builtins.__nuitka_original_args = sys.argv[:]
    sys.argv = [sys.argv[0]]
""",
                "Scrubbing multiprocessing fork arguments from sys.argv.",
            )

        if full_name in ("multiprocessing", "anyio"):
            yield (
                """\
import sys, os
argv0 = sys.argv[0]
if sys.platform == "win32" and not os.path.exists(argv0) and not argv0.endswith(".exe"):
    argv0 += ".exe"

sys.executable = %s
sys._base_executable = sys.executable
""" % ("__nuitka_binary_exe" if isStandaloneMode() else "argv0"),
                """\
Monkey patching "%s" load environment.""" % full_name,
            )

    @staticmethod
    def createPostModuleLoadCode(module):
        full_name = module.getFullName()

        if full_name == "multiprocessing":
            yield (
                """\
try:
    from multiprocessing.forking import ForkingPickler
except ImportError:
    from multiprocessing.reduction import ForkingPickler

import os
import sys

class C:
   def f():
       pass

def _reduce_compiled_method(m):
    if m.im_self is None:
        return getattr, (m.im_class, m.im_func.__name__)
    else:
        return getattr, (m.im_self, m.im_func.__name__)

ForkingPickler.register(type(C().f), _reduce_compiled_method)
if str is bytes:
    ForkingPickler.register(type(C.f), _reduce_compiled_method)

if str is not bytes:
    import multiprocessing.spawn

    if not hasattr(multiprocessing.spawn, "__nuitka_original_fixup_main_from_path"):
        multiprocessing.spawn.__nuitka_original_fixup_main_from_path = multiprocessing.spawn._fixup_main_from_path

        def _fixup_main_from_path_for_nuitka(main_path):
            assert main_path is not None

            import __parents_main__ as parents_main

            sys.modules["__main__"] = parents_main
            sys.modules["__mp_main__"] = parents_main

        multiprocessing.spawn._fixup_main_from_path = _fixup_main_from_path_for_nuitka
""",
                """\
Monkey patching "multiprocessing" for compiled methods.""",
            )

        if full_name == "anyio.to_process":
            yield (
                """\
import anyio.to_process as _anyio_to_process
import os
import pickle
import sys
from importlib.util import module_from_spec, spec_from_file_location

def _process_worker_for_nuitka():
    stdin = sys.stdin
    stdout = sys.stdout
    sys.stdin = open(os.devnull)
    sys.stdout = open(os.devnull, "w")

    stdout.buffer.write(b"READY\\n")
    stdout.buffer.flush()

    while True:
        retval = exception = None

        try:
            payload = pickle.load(stdin.buffer)
            command = payload[0]
            args = payload[1:]
        except EOFError:
            return
        except BaseException as exc:
            exception = exc
        else:
            if command == "run":
                func = args[0]
                func_args = args[1]

                try:
                    retval = func(*func_args)
                except BaseException as exc:
                    exception = exc
            elif command == "init":
                sys.path, main_module_path = args

                del sys.modules["__main__"]

                if main_module_path and os.path.isfile(main_module_path):
                    try:
                        spec = spec_from_file_location("__mp_main__", main_module_path)

                        if spec and spec.loader:
                            main = module_from_spec(spec)
                            spec.loader.exec_module(main)
                            sys.modules["__main__"] = main
                            sys.modules["__mp_main__"] = main
                    except BaseException as exc:
                        exception = exc
                else:
                    try:
                        import __parents_main__ as parents_main
                    except ImportError:
                        pass
                    else:
                        sys.modules["__main__"] = parents_main
                        sys.modules["__mp_main__"] = parents_main

        try:
            if exception is not None:
                status = b"EXCEPTION"
                pickled = pickle.dumps(exception, pickle.HIGHEST_PROTOCOL)
            else:
                status = b"RETURN"
                pickled = pickle.dumps(retval, pickle.HIGHEST_PROTOCOL)
        except BaseException as exc:
            exception = exc
            status = b"EXCEPTION"
            pickled = pickle.dumps(exc, pickle.HIGHEST_PROTOCOL)

        stdout.buffer.write(b"%s %d\\n" % (status, len(pickled)))
        stdout.buffer.write(pickled)
        stdout.buffer.flush()

        if isinstance(exception, SystemExit):
            raise exception

_anyio_to_process.process_worker = _process_worker_for_nuitka
""",
                """\
Monkey patching "anyio.to_process" for worker start.""",
            )

    def createFakeModuleDependency(self, module):
        full_name = module.getFullName()

        if full_name not in ("multiprocessing", "anyio"):
            return

        if (
            self.created_parents_main
            or getModuleInclusionInfoByName("__parents_main__") is not None
        ):
            return

        # First, build the module node and then read again from the
        # source code.
        root_module = getRootTopModule()

        # TODO: Might have to also make "__mp_main__"
        module_name = ModuleName("__parents_main__")

        source_code = readSourceCodeFromFilename(module_name, root_module.getFilename())

        # For the call stack, this may look bad or different to what CPython
        # does, but such is life. The forking module was split up into multiple
        # modules in Python3 so this is doubled. TODO: Move the version check
        # into the code and let Nuitka compile time optimize it, which will also
        # make sure joblib support is there for Python2 as well.
        if python_version >= 0x300:
            source_code += """
def __nuitka_freeze_support():
    import sys
    import builtins
    import multiprocessing.spawn

    # This is a variant of freeze_support that will work for multiprocessing and
    # joblib equally well.
    kwds = {}

    if not hasattr(builtins, "__nuitka_original_args"):
        return

    sys.argv = builtins.__nuitka_original_args
    del builtins.__nuitka_original_args

    for arg in sys.argv[2:]:
        try:
            name, value = arg.split('=')
        except ValueError:
            name = "pipe_handle"
            value = arg

        if value == 'None':
            kwds[name] = None
        else:
            kwds[name] = int(value)

    # Otherwise main module names will not work.
    sys.modules["__main__"] = sys.modules["__parents_main__"]

    multiprocessing.spawn.spawn_main(**kwds)
__nuitka_freeze_support()
"""
        else:
            source_code += """
def __nuitka_freeze_support():
    import __builtin__, sys
    if not hasattr(__builtin__, "__nuitka_original_args"):
        return

    sys.argv = __builtin__.__nuitka_original_args
    del __builtin__.__nuitka_original_args

    sys.modules["__main__"] = sys.modules[__name__]
    __import__("multiprocessing.forking").forking.freeze_support()
__nuitka_freeze_support()
"""

        yield makeFakeModuleDescription(
            module_name=module_name,
            source_code=source_code,
            source_filename=root_module.getCompileTimeFilename(),
            reason="Auto enable multiprocessing/anyio freeze support",
        )

        self.created_parents_main = True

    def onModuleEncounter(
        self, using_module_name, module_name, module_filename, module_kind
    ):
        # Enforce recursion in to multiprocessing for accelerated mode, which
        # would normally avoid this.
        if module_name.hasNamespace("multiprocessing"):
            return True, "Multiprocessing plugin needs this to monkey patch it."

    def decideCompilation(self, module_name):
        if module_name.hasNamespace("multiprocessing"):
            return "bytecode"

        # TODO: Make this demoted too.
        # or module_name in( "multiprocessing-preLoad", "multiprocessing-postLoad"):

    @staticmethod
    def getPreprocessorSymbols():
        if getModuleInclusionInfoByName("__parents_main__"):
            return {"_NUITKA_PLUGIN_MULTIPROCESSING_ENABLED": "1"}


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        https://www.gnu.org/licenses/agpl-3.0.txt
#
#     See also: "Nuitka Runtime Library Exception, Version 1.0" in file
#     "LICENSE-RUNTIME.txt" for additional permissions granted under Section 7.
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#!/usr/bin/python
#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Make version bump for Nuitka."""

import sys

from nuitka.options.CommandLineOptionsTools import makeOptionsParser
from nuitka.tools.Basics import goHome
from nuitka.tools.release.DebianPackage import updateDebianChangelog
from nuitka.tools.release.Release import getBranchName
from nuitka.Tracing import my_print
from nuitka.utils.FileOperations import openTextFile


def getBumpedVersion(mode, old_version):
    if mode == "prerelease":
        if "rc" in old_version:
            parts = old_version.split("rc")

            new_version = "rc".join([parts[0], str(int(parts[1]) + 1)])
        else:
            old_version = ".".join(old_version.split(".")[:3])
            parts = old_version.split(".")
            parts[-1] = str(int(parts[-1]) + 1)

            new_version = ".".join(parts) + "rc1"
    elif mode == "release":
        if "rc" in old_version:
            old_version = old_version[: old_version.find("rc")]
            was_pre = True
        else:
            was_pre = False

        new_version = ".".join(old_version.split(".")[:3])

        if not was_pre:
            parts = new_version.split(".")
            parts[-1] = str(int(parts[-1]) + 1)

            new_version = ".".join(parts)
    elif mode == "hotfix":
        assert "pre" not in old_version and "rc" not in old_version

        parts = old_version.split(".")

        if len(parts) == 2:
            parts.append("1")
        else:
            parts[-1] = str(int(parts[-1]) + 1)

        new_version = ".".join(parts)

    else:
        sys.exit("Error, unknown mode '%s'." % mode)

    return new_version


def main():
    parser = makeOptionsParser(usage=None, epilog=None)

    parser.add_option(
        "--mode",
        action="store",
        dest="mode",
        default=None,
        help="""\
The mode of version number update, "prerelease", "hotfix", "release",
"auto" (default auto determined from branch name), and "re-date" bumps
the changelog date only.""",
    )

    options, positional_args = parser.parse_args()

    if positional_args:
        parser.print_help()

        sys.exit("\nError, no positional argument allowed.")

    # Go its own directory, to have it easy with path knowledge.
    goHome()

    with openTextFile("nuitka/Version.py", "r") as f:
        option_lines = f.readlines()

    (version_line,) = [line for line in option_lines if line.startswith("Nuitka V")]

    old_version = version_line[8:].rstrip()

    mode = options.mode
    branch_name = getBranchName()

    if mode is None:
        if branch_name.startswith("hotfix/"):
            mode = "hotfix"
        elif branch_name == "main" or branch_name.startswith("release/"):
            mode = "release"
        elif branch_name == "develop":
            mode = "prerelease"
        else:
            sys.exit("Error, cannot detect mode from branch name '%s'." % branch_name)

    if mode != "re-date":
        new_version = getBumpedVersion(mode, old_version)
        my_print("Bumped %s '%s' -> '%s'." % (mode, old_version, new_version))

        with openTextFile("nuitka/Version.py", "w") as options_file:
            for line in option_lines:
                if line.startswith("Nuitka V"):
                    line = "Nuitka V" + new_version + "\n"

                options_file.write(line)

        # Debian is currently in not freeze, change to "experimental" once that changes.
        updateDebianChangelog(old_version, new_version, "unstable")
    else:
        # Debian is currently in not freeze, change to "experimental" once that changes.
        updateDebianChangelog(old_version, old_version, "unstable")


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        https://www.gnu.org/licenses/agpl-3.0.txt
#
#     See also: "Nuitka Runtime Library Exception, Version 1.0" in file
#     "LICENSE-RUNTIME.txt" for additional permissions granted under Section 7.
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Generic annotations tests, cover most important forms of them."""

# Tests are dirty on purpose.
#
# pylint: disable=not-an-iterable,unused-argument

import sys
from typing import Tuple, TypeVar, TypeVarTuple

T = TypeVar("T")
Ts = TypeVarTuple("Ts")

print("Module level T=", T)
print("Module level Ts=", Ts)
print()


print("[Part: 1] Unpacking a TypeVar tuple with star list arguments gives:")


def func1(*args: *Ts) -> None:
    print("Function level Ts=", Ts)
    print("Function level *Ts=", *Ts)


func1()
print(func1.__annotations__, "\n")


print("[Part: 2] Manually defining a Tuple[int,...] gives:")


def func21(*args: *Tuple[int, ...]) -> None:
    pass


func21()
print("Annotations", func21.__annotations__, "\n")


def func22(*args: *tuple[int, ...]) -> None:
    pass


func22()
print("Annotations", func22.__annotations__, "\n")


print("[Part: 3] Unpacking a TypeVar tuple with star list arguments gives:")


def func31(*args: Tuple[*Ts]) -> Tuple[*Ts]:
    print("Function level Ts=", Ts)
    print("Function level *Ts=", *Ts)


func31()
print("Annotations", func31.__annotations__, "\n")


def func32(*args: tuple[*Ts]) -> tuple[*Ts]:
    print("Function level Ts=", Ts)
    print("Function level *Ts=", *Ts)


func32()
print("Annotations", func32.__annotations__, "\n")


# TODO: Disabled for Python3.14, because the argument error for print(*T) is not
# yet compatible, but TypeError is raised indeed.
if sys.version_info < (3, 14):
    print(
        "[Part: 4] Unpacking a TypeVar with star list arguments should raise an error:"
    )
    try:

        def func4(*args: *T) -> T:
            print(T)
            print(*T)

        func4()
        print("Annotations", func4.__annotations__, "\n")
    except TypeError as e:
        print("Expected error:", e, "\n")

#     Python tests originally created or extracted from other peoples work. The
#     parts were too small to be protected.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Pack distribution folders into a single file."""

import os
import subprocess
import sys

from nuitka.build.SconsInterface import (
    asBoolStr,
    cleanSconsDirectory,
    getCommonSconsOptions,
    provideStaticSourceFilesOnefile,
    runScons,
)
from nuitka.build.SconsUtils import getSconsReportValue
from nuitka.options.Options import (
    getJobLimit,
    getOnefileTempDirSpec,
    getProgressBar,
    getWindowsSplashScreen,
    isLowMemory,
    isOnefileTempDirMode,
    isRemoveBuildDir,
    shallDisableCompressionCacheUsage,
    shallNotCompressOnefile,
    shallOnefileAsArchive,
    shallTraceExecution,
)
from nuitka.OutputDirectories import getResultFullpath, getSourceDirectoryPath
from nuitka.plugins.Hooks import (
    getBuildDefinitions,
    onBootstrapBinary,
    onGeneratedSourceCode,
    onOnefileFinished,
    writeExtraCodeFiles,
)
from nuitka.PostProcessing import executePostProcessingResources
from nuitka.PythonVersions import (
    getZstandardSupportingVersions,
    python_version,
)
from nuitka.States import states
from nuitka.Tracing import onefile_logger, postprocessing_logger
from nuitka.utils.Execution import withEnvironmentVarsOverridden
from nuitka.utils.FileOperations import (
    areSamePaths,
    getExternalUsePath,
    getFileContents,
    getFileSize,
    makeContainingPath,
    removeDirectory,
)
from nuitka.utils.InstalledPythons import findInstalledPython
from nuitka.utils.Json import writeJsonToFilename
from nuitka.utils.SharedLibraries import cleanupHeaderForAndroid
from nuitka.utils.Signing import addMacOSCodeSignature
from nuitka.utils.Utils import (
    isAndroidBasedLinux,
    isMacOS,
    isWin32OrPosixWindows,
    isWin32Windows,
)
from nuitka.utils.WindowsResources import RT_RCDATA, addResourceToFile

from .DllDependenciesWin32 import shallIncludeWindowsRuntimeDLLs
from .IncludedDataFiles import getIncludedDataFiles
from .IncludedEntryPoints import (
    getStandaloneEntryPoints,
    getStandaloneMainEntryPoint,
)


def packDistFolderToOnefile(dist_dir):
    """Pack distribution to onefile, i.e. a single file that is directly executable."""

    onefile_output_filename = getResultFullpath(onefile=True, real=True)
    start_binary = getResultFullpath(onefile=False, real=True)

    packDistFolderToOnefileBootstrap(
        onefile_output_filename=onefile_output_filename,
        dist_dir=dist_dir,
        start_binary=start_binary,
    )

    onOnefileFinished(onefile_output_filename)


def _runOnefileScons(
    onefile_compression, onefile_archive, backend_resource_mode, onefile_payload_size
):
    scons_options, env_values = getCommonSconsOptions()

    source_dir = getSourceDirectoryPath(onefile=True, create=False)

    # Let plugins do their thing for onefile mode too.
    writeExtraCodeFiles(onefile=True)
    provideStaticSourceFilesOnefile(source_dir=source_dir)
    onGeneratedSourceCode(source_dir=source_dir, onefile=True)

    scons_options["result_exe"] = getResultFullpath(onefile=True, real=False)
    scons_options["source_dir"] = source_dir
    scons_options["debug_mode"] = asBoolStr(states.is_debug)
    scons_options["trace_mode"] = asBoolStr(shallTraceExecution())
    scons_options["onefile_splash_screen"] = asBoolStr(
        getWindowsSplashScreen() is not None
    )
    if isWin32Windows() and shallIncludeWindowsRuntimeDLLs():
        scons_options["onefile_windows_static_runtime"] = asBoolStr(True)

    env_values["_NUITKA_ONEFILE_TEMP_SPEC"] = getOnefileTempDirSpec()
    env_values["_NUITKA_ONEFILE_COMPRESSION_BOOL"] = "1" if onefile_compression else "0"
    env_values["_NUITKA_ONEFILE_ARCHIVE_BOOL"] = "1" if onefile_archive else "0"
    env_values["_NUITKA_ONEFILE_HAS_PAYLOAD_BOOL"] = "1" if hasOnefilePayload() else "0"
    env_values["_NUITKA_ONEFILE_PAYLOAD_SIZE_INT"] = str(onefile_payload_size)

    main_filename_in_payload = hasOnefilePayloadMainEntry()

    if not main_filename_in_payload:
        env_values["_NUITKA_ONEFILE_MAIN_FILENAME"] = (
            getStandaloneMainEntryPoint().dest_path
        )

    # TODO: The None check can go away once all platforms are supported.
    if backend_resource_mode is not None:
        scons_options["backend_resource_mode"] = backend_resource_mode

    # Allow plugins to build definitions.
    env_values.update(getBuildDefinitions())

    result = runScons(
        scons_options=scons_options,
        env_values=env_values,
        scons_filename="Onefile.scons",
    )

    # Exit if compilation failed.
    if not result:
        onefile_logger.sysexit("Error, onefile bootstrap binary build failed.")


_compressor_python = None


def hasOnefilePayload():
    return True


def hasOnefilePayloadMainEntry():
    # Folder mode inherently separates the main entry point from any potential payload
    return "copy" in getStandaloneMainEntryPoint().tags


def getCompressorPython():
    # User may disable it.
    if shallNotCompressOnefile():
        return None

    global _compressor_python  # singleton, pylint: disable=global-statement

    if _compressor_python is None:
        _compressor_python = findInstalledPython(
            python_versions=getZstandardSupportingVersions(),
            module_specs="compression.zstd|zstandard>=0.15",
        )

        if _compressor_python is None:
            if python_version < 0x350:
                onefile_logger.warning("""\
Onefile mode cannot compress without 'zstandard' module installed on \
another discoverable Python >= 3.5 on your system.""")
            else:
                onefile_logger.warning("""\
Onefile mode cannot compress without 'zstandard' package installed You \
probably should depend on 'Nuitka[onefile]' rather than 'Nuitka' which \
among other things depends on it.""")

    return _compressor_python


def runOnefileCompressor(
    compressor_python, dist_dir, onefile_output_filename, start_binary, expected_files
):
    file_checksums = not isOnefileTempDirMode()
    win_path_sep = isWin32OrPosixWindows()

    if compressor_python is None or areSamePaths(
        compressor_python.getPythonExe(), sys.executable
    ):
        from nuitka.tools.onefile_compressor.OnefileCompressor import (
            attachOnefilePayload,
        )

        attachOnefilePayload(
            dist_dir=dist_dir,
            onefile_output_filename=onefile_output_filename,
            start_binary=start_binary,
            include_start_binary=hasOnefilePayloadMainEntry(),
            expect_compression=compressor_python is not None,
            as_archive=shallOnefileAsArchive(),
            use_compression_cache=not shallDisableCompressionCacheUsage(),
            file_checksums=file_checksums,
            win_path_sep=win_path_sep,
            low_memory=isLowMemory(),
            job_limit=getJobLimit(),
            expected_files=expected_files,
        )
    else:
        expected_files_filename = os.path.join(
            getSourceDirectoryPath(onefile=True, create=False), "expected_files.json"
        )
        writeJsonToFilename(expected_files_filename, expected_files)

        onefile_compressor_path = os.path.normpath(
            os.path.join(os.path.dirname(__file__), "..", "tools", "onefile_compressor")
        )

        mapping = {
            "NUITKA_PACKAGE_HOME": os.path.dirname(
                os.path.abspath(sys.modules["nuitka"].__path__[0])
            )
        }

        mapping["NUITKA_PROGRESS_BAR"] = getProgressBar()

        onefile_logger.info(
            "Using external Python '%s' to compress the payload."
            % compressor_python.getPythonExe()
        )

        with withEnvironmentVarsOverridden(mapping):
            subprocess.check_call(
                [
                    compressor_python.getPythonExe(),
                    onefile_compressor_path,
                    dist_dir,
                    getExternalUsePath(onefile_output_filename, only_dirname=True),
                    start_binary,
                    str(hasOnefilePayloadMainEntry()),
                    str(file_checksums),
                    str(win_path_sep),
                    str(isLowMemory()),
                    str(shallOnefileAsArchive()),
                    str(not shallDisableCompressionCacheUsage()),
                    str(getJobLimit()),
                    expected_files_filename,
                ],
                shell=False,
            )


def packDistFolderToOnefileBootstrap(onefile_output_filename, dist_dir, start_binary):
    onefile_logger.info("Creating single file from dist folder, this may take a while.")

    onefile_logger.info("Running bootstrap binary compilation via Scons.")

    # Cleanup first.
    source_dir = getSourceDirectoryPath(onefile=True, create=True)
    cleanSconsDirectory(source_dir)

    # Used only in some configurations
    onefile_payload_filename = os.path.join(source_dir, "blobs", "__payload.bin")
    makeContainingPath(onefile_payload_filename)

    # Now need to append to payload it, potentially compressing it.
    compressor_python = getCompressorPython()

    backend_source_dir = getSourceDirectoryPath(onefile=False, create=False)
    backend_resource_mode = getSconsReportValue(backend_source_dir, "resource_mode")

    # We might not even have a payload due to commercial file embedding.
    has_payload = hasOnefilePayload()
    onefile_payload_size = 0

    if has_payload:
        expected_files = []
        for data_file in getIncludedDataFiles():
            if "copy" in data_file.tags:
                expected_files.append(data_file.dest_path)

        for entry_point in getStandaloneEntryPoints():
            if "copy" in entry_point.tags:
                expected_files.append(entry_point.dest_path)

        expected_files = tuple(expected_files)

        runOnefileCompressor(
            compressor_python=compressor_python,
            dist_dir=dist_dir,
            onefile_output_filename=onefile_payload_filename,
            start_binary=start_binary,
            expected_files=expected_files,
        )

        onefile_payload_size = getFileSize(onefile_payload_filename)

    # Create the bootstrap binary for unpacking.
    _runOnefileScons(
        onefile_compression=compressor_python is not None,
        onefile_archive=shallOnefileAsArchive(),
        backend_resource_mode=backend_resource_mode,
        onefile_payload_size=onefile_payload_size,
    )

    if isWin32Windows():
        executePostProcessingResources(
            result_filename=onefile_output_filename, manifest=None, onefile=True
        )

    if isAndroidBasedLinux():
        cleanupHeaderForAndroid(onefile_output_filename)

    onBootstrapBinary(onefile_output_filename)

    if isMacOS():
        addMacOSCodeSignature(
            filenames=[onefile_output_filename], entitlements_filename=None
        )

    if (
        has_payload
        and getSconsReportValue(source_dir, "resource_mode") == "win_resource"
    ):
        assert isWin32Windows()

        addResourceToFile(
            target_filename=onefile_output_filename,
            data=getFileContents(onefile_payload_filename, mode="rb"),
            resource_kind=RT_RCDATA,
            lang_id=0,
            res_name=27,
            logger=postprocessing_logger,
        )

    if isRemoveBuildDir():
        onefile_logger.info("Removing onefile build directory '%s'." % source_dir)

        removeDirectory(
            path=source_dir,
            logger=onefile_logger,
            ignore_errors=False,
            extra_recommendation=None,
        )
        assert not os.path.exists(source_dir)
    else:
        onefile_logger.info("Keeping onefile build directory '%s'." % source_dir)


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        https://www.gnu.org/licenses/agpl-3.0.txt
#
#     See also: "Nuitka Runtime Library Exception, Version 1.0" in file
#     "LICENSE-RUNTIME.txt" for additional permissions granted under Section 7.
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Walrus dictionary operation corner cases."""


def case1_pop3_unknown_key(key, **kwargs):
    """Walrus with 'kwargs.pop(key, None)' where key hash-ability is unknown."""

    if (value := kwargs.pop(key, None)) is not None:
        return value

    return None


def case2_pop3_union_key(flag, **kwargs):
    """Walrus pop with runtime key being hashable string or unhashable list."""

    key = [] if flag else "x"

    if (value := kwargs.pop(key, None)) is not None:
        return value

    return None


def case3_pop3_nested_walrus(key, **kwargs):
    """Nested walrus pop calls to stress short-circuit ordering and side effects."""

    if (left := kwargs.pop(key, None)) is not None and (
        right := kwargs.pop("y", None)
    ) is not None:
        return left, right

    return None


def case4_setdefault3_unknown_key(key, **kwargs):
    """Walrus with 'kwargs.setdefault(key, 123)' for unknown key hash-ability."""

    if (value := kwargs.setdefault(key, 123)) is not None:
        return value

    return None


def case5_pop3_in_try(key, **kwargs):
    """Walrus pop inside explicit try/except to exercise try node exception tracking."""

    try:
        if (value := kwargs.pop(key, None)) is not None:
            return value
    except TypeError:
        return "caught"

    return None


print("case1-hit", case1_pop3_unknown_key("x", x=1))
print("case1-miss", case1_pop3_unknown_key("x"))

print("case2-false", case2_pop3_union_key(False, x=1))

try:
    print("case2-true", case2_pop3_union_key(True, x=1))
except TypeError:
    print("case2-true", "TypeError")

print("case3-both", case3_pop3_nested_walrus("x", x=1, y=2))
print("case3-missing", case3_pop3_nested_walrus("x", x=1))

print("case4", case4_setdefault3_unknown_key("x"))

print("case5-hit", case5_pop3_in_try("x", x=1))
print("case5-miss", case5_pop3_in_try("x"))
print("case5-caught", case5_pop3_in_try([], x=1))

#     Python tests originally created or extracted from other peoples work. The
#     parts were too small to be protected.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""DLL dependency scan methods that are shared."""

import os

from nuitka.containers.OrderedSets import OrderedSet
from nuitka.importing.Importing import locateModule
from nuitka.plugins.Hooks import getModuleSpecificDllPaths
from nuitka.Tracing import inclusion_logger
from nuitka.utils.FileOperations import (
    getNormalizedPathJoin,
    getSubDirectoriesWithDlls,
)
from nuitka.utils.ModuleNames import ModuleName
from nuitka.utils.SharedLibraries import getWindowsRunningProcessDLLPaths
from nuitka.utils.Utils import getArchitecture

_ld_library_cache = {}
_missing_msvc_redist_dlls = OrderedSet()

# Names of MSVC redistributable DLLs to pick up
# spell-checker: ignore concrt,msvcp,vcruntime,vcamp,vccorlib,codecvt,vcomp
_msvc_redist_dll_names = set(
    (
        "concrt140.dll",
        "msvcp140.dll",
        "msvcp140_1.dll",
        "msvcp140_2.dll",
        "msvcp140_atomic_wait.dll",
        "msvcp140_codecvt_ids.dll",
        "vccorlib140.dll",
        "vcruntime140.dll",
        "vcruntime140_1.dll",
        "vcruntime140_threads.dll",
        "vcamp140.dll",
        "vcomp140.dll",
    )
)


def isMsvcRedistDllName(dll_name):
    return dll_name in _msvc_redist_dll_names


def attemptToFindNotFoundDLL(dll_filename):
    """Some heuristics and tricks to find DLLs not found otherwise."""

    # Lets attempt to find it on currently loaded DLLs, this typically should
    # find the Python DLL.
    currently_loaded_dlls = getWindowsRunningProcessDLLPaths()

    if dll_filename in currently_loaded_dlls:
        return currently_loaded_dlls[dll_filename]

    # Lets try the Windows system, spell-checker: ignore systemroot
    dll_filename = getNormalizedPathJoin(
        os.environ["SYSTEMROOT"],
        "SysWOW64" if getArchitecture() == "x86" else "System32",
        dll_filename,
    )
    dll_filename = os.path.normcase(dll_filename)

    if os.path.exists(dll_filename):
        return dll_filename

    return None


def addMissingMsvcRedistDLL(dll_name):
    _missing_msvc_redist_dlls.add(dll_name)


def reportMissingMsvcRedistDLLs():
    if _missing_msvc_redist_dlls:
        inclusion_logger.warning(
            """\
The following Visual C++ Redistributable DLLs were not found: %s. \
For a fully portable standalone distribution, these DLLs must be \
available either by installing the Microsoft Visual C++ Redistributable \
for Visual Studio 2015-2022 on the target system or by bundling them with \
the application. To bundle them, Visual Studio must be installed on the build machine."""
            % ", ".join(sorted(_missing_msvc_redist_dlls))
        )
        _missing_msvc_redist_dlls.clear()


def shallIgnoreMissingDLL(dll_name):
    if dll_name.startswith(("api-ms-win-", "ext-ms-win-")):
        return True

    # Ignore this runtime DLL of Python2, will be coming via manifest.
    # spell-checker: ignore msvcr90
    if dll_name == "msvcr90.dll":
        return True

    # Ignore UCRT runtime, this must come from OS, spell-checker: ignore ucrtbase
    if dll_name == "ucrtbase.dll":
        return True

    return False


def getLdLibraryPath(package_name, python_rpaths, original_dir):
    key = package_name, tuple(python_rpaths), original_dir

    if key not in _ld_library_cache:
        ld_library_path = OrderedSet()
        if python_rpaths:
            ld_library_path.update(python_rpaths)

        ld_library_path.update(
            getPackageSpecificDLLDirectories(
                package_name=package_name,
                consider_plugins=True,
            )
        )
        if original_dir is not None:
            ld_library_path.add(original_dir)

        _ld_library_cache[key] = ld_library_path

    return _ld_library_cache[key]


def getPackageSpecificDLLDirectories(
    package_name, consider_plugins, allow_not_found=False
):
    scan_dirs = OrderedSet()

    if package_name is not None:
        package_dir = locateModule(
            module_name=package_name, parent_package=None, level=0
        )[1]

        if package_dir is None:
            if allow_not_found:
                return scan_dirs

            return inclusion_logger.sysexit("""\
Error, failed to locate package '%s' while trying to look up DLL dependencies, \
that should not happen. Please report the issue.""" % package_name)

        if os.path.isdir(package_dir):
            scan_dirs.add(package_dir)
            scan_dirs.update(getSubDirectoriesWithDlls(package_dir))

        if consider_plugins:
            for plugin_provided_dir in getModuleSpecificDllPaths(package_name):
                if os.path.isdir(plugin_provided_dir):
                    scan_dirs.add(plugin_provided_dir)
                    scan_dirs.update(getSubDirectoriesWithDlls(plugin_provided_dir))

    # TODO: Move this to plugins DLLs section.
    if package_name == "torchvision" and consider_plugins:
        scan_dirs.update(
            getPackageSpecificDLLDirectories(
                package_name=ModuleName("torch"),
                consider_plugins=True,
                allow_not_found=True,
            )
        )

    return scan_dirs


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        https://www.gnu.org/licenses/agpl-3.0.txt
#
#     See also: "Nuitka Runtime Library Exception, Version 1.0" in file
#     "LICENSE-RUNTIME.txt" for additional permissions granted under Section 7.
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#!/usr/bin/python
#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Coverage rendering tool."""

import os
import shutil
import subprocess
import sys


def main():
    # Set the stage.
    os.chdir(os.path.dirname(__file__))
    shutil.rmtree("coverage", ignore_errors=True)

    print("Fetching coverage files:")

    subprocess.call(
        ["rsync", "-az", "--delete", "%s/" % os.environ["COVERAGE_DIR"], "coverage/"]
    )

    print("Combining coverage files:")
    os.chdir("coverage")

    print("Detect coverage file roots:")
    # Now detect where the files were collected from:

    paths = [os.path.abspath(os.path.join(os.curdir, "..", ".."))]

    for filename in os.listdir("."):
        if not filename.startswith("meta.coverage"):
            continue

        values = {}
        with open(filename, encoding="utf-8") as f:
            # pylint: disable=exec-used
            exec(f.read(), values)

        if "__builtins__" in values:
            del values["__builtins__"]

        paths.append(values["NUITKA_SOURCE_DIR"])

    # spell-checker: ignore coveragerc
    coverage_path = os.path.abspath(".coveragerc")

    with open(coverage_path, "w", encoding="utf-8") as coverage_rc_file:
        coverage_rc_file.write("[paths]\n")
        coverage_rc_file.write("source = \n")

        for path in paths:
            coverage_rc_file.write("   " + path + "\n")

    subprocess.call(
        # spell-checker: ignore rcfile
        [
            sys.executable,
            "-m",
            "coverage",
            "combine",
            "--rcfile",
            coverage_path,
        ]
    )

    assert os.path.exists(coverage_path)

    subprocess.call(
        [
            sys.executable,
            "-m",
            "coverage",
            "html",
            "--rcfile",
            coverage_path,
        ]
    )

    # Clean up after ourselves again.
    # shutil.rmtree("coverage", ignore_errors = True)


if __name__ == "__main__":
    main()

#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        https://www.gnu.org/licenses/agpl-3.0.txt
#
#     See also: "Nuitka Runtime Library Exception, Version 1.0" in file
#     "LICENSE-RUNTIME.txt" for additional permissions granted under Section 7.
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#!/usr/bin/env python
#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Runner for Python PGO tests of Nuitka.

PGO tests attempt to cover inclusion/non-inclusion of code and detection of
types.

"""

import os
import sys

# Find nuitka package relative to us.
sys.path.insert(
    0,
    os.path.normpath(
        os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", "..")
    ),
)

# isort:start

from nuitka.reports.CompilationReportReader import (
    extractModulesUsedByModule,
    parseCompilationReport,
)
from nuitka.tools.testing.Common import (
    compareWithCPython,
    createSearchMode,
    scanDirectoryForTestCaseFolders,
    setup,
)


def main():
    setup(suite="pgo", needs_io_encoding=True)

    search_mode = createSearchMode()

    # Now run all the tests in this directory.
    for filename, filename_main in scanDirectoryForTestCaseFolders("."):
        active = search_mode.consider(dirname=None, filename=filename)

        if active:
            report_filename = "compilation-report-%s.xml" % filename

            extra_flags = [
                # No error exits normally, unless we break tests, and that we would
                # like to know.
                "expect_success",
                # Keep no temporary files.
                "remove_output",
                # Include imported files, PGO will then have to deal with unused ones.
                "--follow-imports",
                # The output during compilation from PGO capture is harmful, so
                # split compilation and execution of final result.
                "two_step_execution",
                # Inclusion report is used by the testing of expected things included
                # or not.
                "--report=%s" % report_filename,
                # Cache the CPython results for reuse, they will normally not change.
                "cpython_cache",
            ]

            compareWithCPython(
                dirname=filename,
                filename=filename_main,
                extra_flags=extra_flags,
                search_mode=search_mode,
            )

            compilation_report = parseCompilationReport(report_filename)

            modules_used = extractModulesUsedByModule(
                compilation_report=compilation_report, module_name="__main__"
            )
            assert (
                modules_used["ImportedButMaybeNotUsed"]["exclusion_reason"]
                == "PGO based decision"
            ), modules_used

    search_mode.finish()


if __name__ == "__main__":
    main()

#     Python test originally created or extracted from other peoples work. The
#     parts from me are licensed as below. It is at least Free Software where
#     it's copied from other people. In these cases, that will normally be
#     indicated.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


import itertools

module_value1 = module_value2 = module_value3 = module_value4 = 1000
module_key1 = module_key2 = module_key3 = module_key4 = 1000


def calledRepeatedly(cond):
    # Force frame and eliminate forward propagation (currently), and use local
    # variables to avoid impact of global variable access.
    dict_key1 = module_value1
    dict_key2 = module_value2
    dict_key3 = module_value3
    dict_key4 = module_value4

    dict_val1 = module_value1
    dict_val2 = module_value2
    dict_val3 = module_value3
    dict_val4 = module_value4

    l = 1

    if cond:
        l = {
            dict_key1: dict_val1,
            dict_key2: dict_val2,
            dict_key3: dict_val3,
            dict_key4: dict_val4,
        }

    return (
        l,
        dict_val1,
        dict_val2,
        dict_val3,
        dict_val4,
        dict_key1,
        dict_key2,
        dict_key3,
        dict_key4,
    )


for x in itertools.repeat(None, 10000):
    # construct_begin
    calledRepeatedly(True)
    # construct_alternative
    calledRepeatedly(False)
    # construct_end

print("OK.")

#     Python test originally created or extracted from other peoples work. The
#     parts from me are licensed as below. It is at least Free Software where
#     it's copied from other people. In these cases, that will normally be
#     indicated.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


# nuitka-project: --mode=standalone

# The issue this is trying to solve only affects Python3.11, and is that due a
# bug in the bytecode attribute caching, compiled modules have to mimick very
# precisely what indexes in dictionaries of modules are used for the dunder
# values.

# nuitka-skip-unless-expression: sys.version_info[:2] == (3, 11)

from __future__ import print_function

import inspect

import some_module
import some_package


def getKeyIndex(module, key):
    return list(module.__dict__).index(key)


if "__compiled__" in globals():
    compiled_module = some_module
    compiled_package = some_package

    # This is a targeted regression for compiled module/package alignment.
    compiled_module_builtins_index = getKeyIndex(compiled_module, "__builtins__")
    compiled_package_builtins_index = getKeyIndex(compiled_package, "__builtins__")
    assert compiled_module_builtins_index == compiled_package_builtins_index, (
        compiled_module_builtins_index,
        compiled_package_builtins_index,
        list(compiled_module.__dict__),
        list(compiled_package.__dict__),
    )

    compiled_module_file_index = getKeyIndex(compiled_module, "__file__")
    compiled_package_file_index = getKeyIndex(compiled_package, "__file__")
    assert compiled_module_file_index == compiled_package_file_index, (
        compiled_module_file_index,
        compiled_package_file_index,
        list(compiled_module.__dict__),
        list(compiled_package.__dict__),
    )

    for _unused in range(32):
        inspect.getfile(compiled_module)

    package_filename = inspect.getfile(compiled_package)
    assert package_filename.endswith("__init__.py"), repr(package_filename)

print("OK")

#     Python test originally created or extracted from other peoples work. The
#     parts from me are licensed as below. It is at least Free Software where
#     it's copied from other people. In these cases, that will normally be
#     indicated.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Functions to handle git staged content.

Inspired from https://raw.githubusercontent.com/hallettj/git-format-staged/master/git-format-staged
Original author: Jesse Hallett <jesse@sitr.us>

spell-checker: ignore Hallett,unpushed
"""

import os
import re

from nuitka.containers.OrderedSets import OrderedSet
from nuitka.format.FileFormatting import cleanupWindowsNewlines
from nuitka.tools.Basics import goHome
from nuitka.Tracing import my_print, tools_logger
from nuitka.utils.CStrings import decodeCStringToPython
from nuitka.utils.Execution import (
    NuitkaCalledProcessError,
    check_call,
    check_output,
    executeProcess,
)
from nuitka.utils.FileOperations import openTextFile


# Parse output from `git diff-index`
def _parseIndexDiffLine(line):
    """Parse output from `git diff-index` into a dictionary."""
    pattern = re.compile(
        r"^:(\d+) (\d+) ([a-f0-9]+) ([a-f0-9]+) ([A-Z])(\d+)?\t([^\t]+)(?:\t([^\t]+))?$"
    )

    zeroed_pat = re.compile("^0+$")

    # Returns the argument unless the argument is a string of zeroes, in which case
    # returns `None`
    def unless_zeroed(s):
        return s if not zeroed_pat.match(s) else None

    match = pattern.match(line)
    if not match:
        raise ValueError("Failed to parse diff-index line: " + line)

    def parseGitPath(value):
        if value is None:
            return None

        if value.startswith('"'):
            return decodeCStringToPython(value).decode("utf8")

        return value

    return {
        "src_mode": unless_zeroed(match.group(1)),
        "dst_mode": unless_zeroed(match.group(2)),
        "src_hash": unless_zeroed(match.group(3)),
        "dst_hash": unless_zeroed(match.group(4)),
        "status": match.group(5),
        "score": int(match.group(6)) if match.group(6) else None,
        "src_path": parseGitPath(match.group(7)),
        "dst_path": parseGitPath(match.group(8)),
    }


def getCheckoutFileChangeDesc(staged):
    """Get descriptions of changed files in the checkout.

    Args:
        staged: bool - if True, look at staged changes (--cached),
                       otherwise look at unstaged changes.
    """
    # Only file additions and modifications
    command = ["git", "diff-index", "--diff-filter=AM", "--no-renames"]

    if staged:
        command.append("--cached")

    command.append("HEAD")

    output = check_output(command)

    for line in output.splitlines():
        if str is not bytes:
            line = line.decode("utf8")

        yield _parseIndexDiffLine(line)


def getModifiedPaths():
    """Get a list of all modified paths in the repository."""
    result = set()

    output = check_output(["git", "diff", "--name-only"])

    for line in output.splitlines():
        if str is not bytes:
            line = line.decode("utf8")

        result.add(line)

    output = check_output(["git", "diff", "--cached", "--name-only"])

    for line in output.splitlines():
        if str is not bytes:
            line = line.decode("utf8")

        result.add(line)

    return tuple(sorted(filename for filename in result if os.path.exists(filename)))


def getRemoteURL(remote_name):
    """Get the URL of a git remote."""
    output = check_output(["git", "remote", "get-url", remote_name])

    if str is not bytes:
        output = output.decode("utf8")

    return output.strip()


def getCurrentBranchName():
    """Get the name of the current git branch."""
    try:
        output = check_output(["git", "branch", "--show-current"])
    except NuitkaCalledProcessError:
        output = check_output(["git", "symbolic-ref", "--short", "HEAD"])

    if str is not bytes:
        output = output.decode("utf8")

    return output.strip()


def getNotPushedPaths():
    """Get a list of modified paths that have not been pushed to upstream."""
    result = set()

    try:
        output = check_output(
            [
                "git",
                "diff",
                "--stat",
                "--name-only",
                "--ignore-submodules=all",
                "@{upstream}",
            ]
        )
    except NuitkaCalledProcessError:
        return result

    for line in output.splitlines():
        if str is not bytes:
            line = line.decode("utf8")

        # Removed files appear too, but are useless to talk about.
        if not os.path.exists(line):
            continue

        result.add(line)

    return tuple(sorted(result))


def getFileHashContent(object_hash):
    """Get the content of a git object from its hash."""
    return check_output(["git", "cat-file", "-p", object_hash])


def putFileHashContent(filename):
    """Add a file's content to the git object database and return its hash."""
    with openTextFile(filename, "r") as input_file:
        new_hash = check_output(
            ["git", "hash-object", "-w", "--stdin"], stdin=input_file
        )

    if str is not bytes:
        new_hash = new_hash.decode("utf8")

    assert new_hash
    return new_hash.rstrip()


def updateFileIndex(diff_entry, new_object_hash):
    """Update the git index with a new hash for a file."""
    # spell-checker: ignore cacheinfo
    check_call(
        [
            "git",
            "update-index",
            "--cacheinfo",
            "%s,%s,%s"
            % (diff_entry["dst_mode"], new_object_hash, diff_entry["src_path"]),
        ]
    )


def updateGitFile(path, orig_object_hash, new_object_hash, staged):
    """Apply a patch to a file in the git repository.

    Args:
        path: str - path to the file
        orig_object_hash: str - original hash of the file
        new_object_hash: str - new hash of the file
        staged: bool - if True, apply as a staged change
    """
    patch = check_output(
        ["git", "diff", "--no-color", orig_object_hash, new_object_hash]
    )

    git_path = path.replace(os.path.sep, "/").encode("utf8")

    def updateLine(line):
        if line.startswith(b"diff --git"):
            line = b"diff --git a/%s b/%s" % (git_path, git_path)
        elif line.startswith(b"--- a/"):
            line = b"--- a/" + git_path
        elif line.startswith(b"+++ b/"):
            line = b"+++ b/" + git_path

        return line

    # Substitute object hashes in patch header with path to working tree file
    patch = b"\n".join(updateLine(line) for line in patch.splitlines()) + b"\n"

    command = ["git", "apply"]

    if not staged:
        command.append("--cached")

    command.append("-")

    # Apply the patch.
    process_result = executeProcess(
        command,
        stdin=patch,
    )

    if process_result.exit_code != 0 and os.name == "nt":
        cleanupWindowsNewlines(path, path)

        process_result = executeProcess(
            ["git", "apply", "-"],
            stdin=patch,
        )

    success = process_result.exit_code == 0

    if not success:
        # TODO: In case of failure, do we need to abort, or what do we do.

        if process_result.stdout:
            my_print(process_result.stdout, style="yellow")
        if process_result.stderr:
            my_print(process_result.stderr, style="yellow")

    return success


def addGitArguments(parser, verb="Analyze"):
    parser.add_option(
        "--diff",
        action="store_true",
        dest="diff",
        default=False,
        help="""\
%s the changed files in git checkout. Default is %%default.""" % verb,
    )

    parser.add_option(
        "--un-pushed",
        "--unpushed",
        action="store_true",
        dest="un_pushed",
        default=False,
        help="""\
%s the changed files in git not yet pushed. Default is %%default.""" % verb,
    )


def getGitPaths(options, positional_args, default_positional_args):
    if options.diff or options.un_pushed:
        if positional_args:
            tools_logger.sysexit(
                "Error, no filenames argument allowed in git diff mode."
            )

        goHome()

        result = OrderedSet()
        if options.diff:
            result.update(getModifiedPaths())

        if options.un_pushed:
            result.update(getNotPushedPaths())
    else:
        result = positional_args

        if not result:
            goHome()
            result = default_positional_args

    return result


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        https://www.gnu.org/licenses/agpl-3.0.txt
#
#     See also: "Nuitka Runtime Library Exception, Version 1.0" in file
#     "LICENSE-RUNTIME.txt" for additional permissions granted under Section 7.
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Shared constants blob format specification handling."""

import os
import re

from nuitka.__past__ import to_byte
from nuitka.utils.FileOperations import getFileContents, getNormalizedPathJoin

_define_pattern = re.compile(
    r"^\s*#define\s+(NUITKA_CONSTANT_BLOB_[A-Z0-9_]+)\s+([0-9A-Fa-fxXuUlL]+)\b"
)

_tag_define_names = (
    ("previous", "NUITKA_CONSTANT_BLOB_TAG_PREVIOUS"),
    ("none", "NUITKA_CONSTANT_BLOB_TAG_NONE"),
    ("true", "NUITKA_CONSTANT_BLOB_TAG_TRUE"),
    ("false", "NUITKA_CONSTANT_BLOB_TAG_FALSE"),
    ("tuple", "NUITKA_CONSTANT_BLOB_TAG_TUPLE"),
    ("list", "NUITKA_CONSTANT_BLOB_TAG_LIST"),
    ("dict", "NUITKA_CONSTANT_BLOB_TAG_DICT"),
    ("set", "NUITKA_CONSTANT_BLOB_TAG_SET"),
    ("frozenset", "NUITKA_CONSTANT_BLOB_TAG_FROZENSET"),
    ("long_positive_small", "NUITKA_CONSTANT_BLOB_TAG_LONG_POSITIVE_SMALL"),
    ("long_negative_small", "NUITKA_CONSTANT_BLOB_TAG_LONG_NEGATIVE_SMALL"),
    ("long_positive_large", "NUITKA_CONSTANT_BLOB_TAG_LONG_POSITIVE_LARGE"),
    ("long_negative_large", "NUITKA_CONSTANT_BLOB_TAG_LONG_NEGATIVE_LARGE"),
    ("int_positive", "NUITKA_CONSTANT_BLOB_TAG_INT_POSITIVE"),
    ("int_negative", "NUITKA_CONSTANT_BLOB_TAG_INT_NEGATIVE"),
    ("float_special", "NUITKA_CONSTANT_BLOB_TAG_FLOAT_SPECIAL"),
    ("float", "NUITKA_CONSTANT_BLOB_TAG_FLOAT"),
    ("text_empty", "NUITKA_CONSTANT_BLOB_TAG_TEXT_EMPTY"),
    ("text_single", "NUITKA_CONSTANT_BLOB_TAG_TEXT_SINGLE"),
    (
        "text_utf8_length_prefixed",
        "NUITKA_CONSTANT_BLOB_TAG_TEXT_UTF8_LENGTH_PREFIXED",
    ),
    (
        "text_utf8_zero_terminated",
        "NUITKA_CONSTANT_BLOB_TAG_TEXT_UTF8_ZERO_TERMINATED",
    ),
    ("attribute_name", "NUITKA_CONSTANT_BLOB_TAG_ATTRIBUTE_NAME"),
    ("bytes_length_prefixed", "NUITKA_CONSTANT_BLOB_TAG_BYTES_LENGTH_PREFIXED"),
    ("bytes_zero_terminated", "NUITKA_CONSTANT_BLOB_TAG_BYTES_ZERO_TERMINATED"),
    ("bytes_single", "NUITKA_CONSTANT_BLOB_TAG_BYTES_SINGLE"),
    ("slice", "NUITKA_CONSTANT_BLOB_TAG_SLICE"),
    ("range", "NUITKA_CONSTANT_BLOB_TAG_RANGE"),
    ("complex_special", "NUITKA_CONSTANT_BLOB_TAG_COMPLEX_SPECIAL"),
    ("complex", "NUITKA_CONSTANT_BLOB_TAG_COMPLEX"),
    ("bytearray", "NUITKA_CONSTANT_BLOB_TAG_BYTEARRAY"),
    ("builtin_anon", "NUITKA_CONSTANT_BLOB_TAG_BUILTIN_ANON"),
    ("builtin_special", "NUITKA_CONSTANT_BLOB_TAG_BUILTIN_SPECIAL"),
    ("blob_data", "NUITKA_CONSTANT_BLOB_TAG_BLOB_DATA"),
    ("generic_alias", "NUITKA_CONSTANT_BLOB_TAG_GENERIC_ALIAS"),
    ("union_type", "NUITKA_CONSTANT_BLOB_TAG_UNION_TYPE"),
    ("builtin_named", "NUITKA_CONSTANT_BLOB_TAG_BUILTIN_NAMED"),
    ("builtin_exception", "NUITKA_CONSTANT_BLOB_TAG_BUILTIN_EXCEPTION"),
    ("code_object", "NUITKA_CONSTANT_BLOB_TAG_CODE_OBJECT"),
    ("end", "NUITKA_CONSTANT_BLOB_TAG_END"),
)

_float_special_define_names = (
    ("pos_zero", "NUITKA_CONSTANT_BLOB_FLOAT_SPECIAL_POS_ZERO"),
    ("neg_zero", "NUITKA_CONSTANT_BLOB_FLOAT_SPECIAL_NEG_ZERO"),
    ("pos_nan", "NUITKA_CONSTANT_BLOB_FLOAT_SPECIAL_POS_NAN"),
    ("neg_nan", "NUITKA_CONSTANT_BLOB_FLOAT_SPECIAL_NEG_NAN"),
    ("pos_inf", "NUITKA_CONSTANT_BLOB_FLOAT_SPECIAL_POS_INF"),
    ("neg_inf", "NUITKA_CONSTANT_BLOB_FLOAT_SPECIAL_NEG_INF"),
)

_code_flag_define_names = (
    ("code_flag_qualname", "NUITKA_CONSTANT_BLOB_CODE_FLAG_QUALNAME"),
    ("code_flag_free_vars", "NUITKA_CONSTANT_BLOB_CODE_FLAG_FREE_VARS"),
    ("code_flag_kw_only", "NUITKA_CONSTANT_BLOB_CODE_FLAG_KW_ONLY"),
    ("code_flag_pos_only", "NUITKA_CONSTANT_BLOB_CODE_FLAG_POS_ONLY"),
    ("code_flag_optimized", "NUITKA_CONSTANT_BLOB_CODE_FLAG_OPTIMIZED"),
    ("code_flag_newlocals", "NUITKA_CONSTANT_BLOB_CODE_FLAG_NEWLOCALS"),
    ("code_flag_varargs", "NUITKA_CONSTANT_BLOB_CODE_FLAG_VARARGS"),
    ("code_flag_varkeywords", "NUITKA_CONSTANT_BLOB_CODE_FLAG_VARKEYWORDS"),
    (
        "code_flag_future_division",
        "NUITKA_CONSTANT_BLOB_CODE_FLAG_FUTURE_DIVISION",
    ),
    (
        "code_flag_future_unicode_literals",
        "NUITKA_CONSTANT_BLOB_CODE_FLAG_FUTURE_UNICODE_LITERALS",
    ),
    (
        "code_flag_future_print_function",
        "NUITKA_CONSTANT_BLOB_CODE_FLAG_FUTURE_PRINT_FUNCTION",
    ),
    (
        "code_flag_future_absolute_import",
        "NUITKA_CONSTANT_BLOB_CODE_FLAG_FUTURE_ABSOLUTE_IMPORT",
    ),
    (
        "code_flag_future_generator_stop",
        "NUITKA_CONSTANT_BLOB_CODE_FLAG_FUTURE_GENERATOR_STOP",
    ),
    (
        "code_flag_future_annotations",
        "NUITKA_CONSTANT_BLOB_CODE_FLAG_FUTURE_ANNOTATIONS",
    ),
    (
        "code_flag_future_barry_as_bdfl",
        "NUITKA_CONSTANT_BLOB_CODE_FLAG_FUTURE_BARRY_AS_BDFL",
    ),
    ("code_flag_nofree", "NUITKA_CONSTANT_BLOB_CODE_FLAG_NOFREE"),
)

_code_kind_define_names = (
    ("code_kind_mask", "NUITKA_CONSTANT_BLOB_CODE_KIND_MASK"),
    ("code_kind_generator", "NUITKA_CONSTANT_BLOB_CODE_KIND_GENERATOR"),
    ("code_kind_coroutine", "NUITKA_CONSTANT_BLOB_CODE_KIND_COROUTINE"),
    ("code_kind_asyncgen", "NUITKA_CONSTANT_BLOB_CODE_KIND_ASYNCGEN"),
)

_required_define_names = (
    ("tags", _tag_define_names),
    ("float_specials", _float_special_define_names),
    ("code_flags", _code_flag_define_names),
    ("code_kinds", _code_kind_define_names),
)

_constant_blob_spec_cache = {}


class _ConstantBlobSpec(object):
    def __init__(self, filename, values):
        self.filename = filename

        for field_name, define_name in _tag_define_names:
            setattr(self, "tag_" + field_name, to_byte(values[define_name]))

        for field_name, define_name in _float_special_define_names:
            setattr(self, "float_special_" + field_name, values[define_name])

        for field_name, define_name in _code_flag_define_names:
            setattr(self, field_name, values[define_name])

        for field_name, define_name in _code_kind_define_names:
            setattr(self, field_name, values[define_name])


def _getDefaultConstantBlobSpecFilename():
    return getNormalizedPathJoin(
        os.path.dirname(__file__), "include", "nuitka", "constants_blob_spec.h"
    )


def _getActiveConstantBlobSpecFilename():
    filename = os.getenv("NUITKA_CONSTANTS_BLOB_HEADER")

    if filename is None:
        filename = _getDefaultConstantBlobSpecFilename()

    return os.path.normpath(os.path.abspath(filename))


def _parseConstantBlobSpecValue(value):
    value = value.rstrip("uUlL")

    return int(value, 0)


def _isIgnorableConstantBlobSpecLine(line):
    line = line.strip()

    if not line:
        return True

    if line in (
        "#ifndef __NUITKA_CONSTANTS_BLOB_SPEC_H__",
        "#define __NUITKA_CONSTANTS_BLOB_SPEC_H__",
        "#endif",
    ):
        return True

    if (
        line.startswith("//")
        or line.startswith("/*")
        or line.startswith("*")
        or line == "*/"
    ):
        return True

    return False


def _loadConstantBlobSpecValues(filename, logger):
    result = {}

    for count, line in enumerate(
        getFileContents(filename, encoding="ascii").splitlines(), 1
    ):
        match = _define_pattern.match(line)

        if match is None:
            if _isIgnorableConstantBlobSpecLine(line):
                continue

            return logger.sysexit(
                "Unexpected line %d in constants blob spec '%s': %r"
                % (count, filename, line)
            )

        define_name, define_value = match.groups()
        result[define_name] = _parseConstantBlobSpecValue(define_value)

    return result


def _checkRequiredConstantBlobSpecValues(filename, values, logger):
    missing = []

    for _kind, define_names in _required_define_names:
        if type(define_names) is tuple:
            for _field_name, define_name in define_names:
                if define_name not in values:
                    missing.append(define_name)
        elif define_names not in values:
            missing.append(define_names)

    if missing:
        return logger.sysexit(
            "Missing constants blob spec values in '%s': %s"
            % (filename, ", ".join(sorted(missing)))
        )


def _checkConstantBlobTagValues(filename, values, logger):
    used_tag_values = set()

    for _field_name, define_name in _tag_define_names:
        tag_value = values[define_name]

        if tag_value <= 0 or tag_value >= 256:
            return logger.sysexit(
                "Invalid constants blob tag value in '%s' for %s: %r"
                % (filename, define_name, tag_value)
            )

        if tag_value in used_tag_values:
            return logger.sysexit(
                "Duplicate constants blob tag value in '%s' for %s: %r"
                % (filename, define_name, tag_value)
            )

        used_tag_values.add(tag_value)


def _checkConstantBlobFloatSpecialValues(filename, values, logger):
    used_float_special_values = set()

    for _field_name, define_name in _float_special_define_names:
        special_value = values[define_name]

        if special_value < 0 or special_value >= 256:
            return logger.sysexit(
                "Invalid float special value in '%s' for %s: %r"
                % (filename, define_name, special_value)
            )

        if special_value in used_float_special_values:
            return logger.sysexit(
                "Duplicate float special value in '%s' for %s: %r"
                % (filename, define_name, special_value)
            )

        used_float_special_values.add(special_value)


def _checkConstantBlobSingleCodeFlagValues(filename, values, logger):
    used_code_flag_values = 0

    for _field_name, define_name in _code_flag_define_names:
        code_flag_value = values[define_name]

        if code_flag_value <= 0 or code_flag_value >= (1 << 64):
            return logger.sysexit(
                "Invalid code flag value in '%s' for %s: %r"
                % (filename, define_name, code_flag_value)
            )

        if code_flag_value & (code_flag_value - 1):
            return logger.sysexit(
                "Non-bitmask code flag value in '%s' for %s: %r"
                % (filename, define_name, code_flag_value)
            )

        if used_code_flag_values & code_flag_value:
            return logger.sysexit(
                "Duplicate code flag value in '%s' for %s: %r"
                % (filename, define_name, code_flag_value)
            )

        used_code_flag_values |= code_flag_value

    return used_code_flag_values


def _checkConstantBlobCodeKindValues(filename, values, used_code_flag_values, logger):
    code_kind_mask = values["NUITKA_CONSTANT_BLOB_CODE_KIND_MASK"]

    if code_kind_mask <= 0 or code_kind_mask >= (1 << 64):
        return logger.sysexit(
            "Invalid code kind mask in '%s': %r" % (filename, code_kind_mask)
        )

    if used_code_flag_values & code_kind_mask:
        return logger.sysexit("Overlapping code flag values in '%s'" % filename)

    used_code_kind_values = set()

    for _field_name, define_name in _code_kind_define_names[1:]:
        code_kind_value = values[define_name]

        if code_kind_value <= 0 or code_kind_value >= (1 << 64):
            return logger.sysexit(
                "Invalid code kind value in '%s' for %s: %r"
                % (filename, define_name, code_kind_value)
            )

        if code_kind_value & ~code_kind_mask:
            return logger.sysexit(
                "Code kind value outside mask in '%s' for %s: %r"
                % (filename, define_name, code_kind_value)
            )

        if code_kind_value in used_code_kind_values:
            return logger.sysexit(
                "Duplicate code kind value in '%s' for %s: %r"
                % (filename, define_name, code_kind_value)
            )

        used_code_kind_values.add(code_kind_value)


def _checkConstantBlobCodeFlagValues(filename, values, logger):
    used_code_flag_values = _checkConstantBlobSingleCodeFlagValues(
        filename=filename, values=values, logger=logger
    )

    _checkConstantBlobCodeKindValues(
        filename=filename,
        values=values,
        used_code_flag_values=used_code_flag_values,
        logger=logger,
    )


def _checkConstantBlobSpecValues(filename, values, logger):
    _checkRequiredConstantBlobSpecValues(
        filename=filename, values=values, logger=logger
    )
    _checkConstantBlobTagValues(filename=filename, values=values, logger=logger)
    _checkConstantBlobFloatSpecialValues(
        filename=filename, values=values, logger=logger
    )
    _checkConstantBlobCodeFlagValues(filename=filename, values=values, logger=logger)


def loadConstantBlobSpec(logger):
    filename = _getActiveConstantBlobSpecFilename()

    result = _constant_blob_spec_cache.get(filename)

    if result is None:
        values = _loadConstantBlobSpecValues(filename, logger)

        _checkConstantBlobSpecValues(filename, values, logger)

        result = _ConstantBlobSpec(filename=filename, values=values)
        _constant_blob_spec_cache[filename] = result

    return result


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        https://www.gnu.org/licenses/agpl-3.0.txt
#
#     See also: "Nuitka Runtime Library Exception, Version 1.0" in file
#     "LICENSE-RUNTIME.txt" for additional permissions granted under Section 7.
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
